#include "RTDS_TTCN_SUT_interface.h"
#include "SIP_TypesAndConf.h"

#include "RTDS_TTCN_MACRO.h"
#include "RTDS_TTCN.h"
#include "RTDS_TTCN_gen.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <time.h>

#ifdef WIN32
#include <windows.h>
#include <winsock2.h>
#else
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#endif

#include "sofia-sip/msg.h"
#include "sofia-sip/sip.h"
#include "sofia-sip/sip_header.h"
#include "sofia-sip/sip_extra.h"


#define BUFFER_SIZE 8192

#define DEFAULT_SUT_ADDRESS "127.0.0.1"
#define DEFAULT_SUT_PORT 5060 /* TTCN_constant_DEFAULT_SIP_PORT */

#ifdef WIN32
#define IS_INVALID_SOCKET(_SOCK) (_SOCK == INVALID_SOCKET)
#define IS_SOCKET_ERROR(_SOCK) (_SOCK == SOCKET_ERROR)
#define SOCKET_ERROR_NUMBER WSAGetLastError()
#define THREAD HANDLE
#else
#define SOCKET int
#define closesocket close
#define IS_INVALID_SOCKET(_SOCK) (_SOCK < 0)
#define IS_SOCKET_ERROR(_SOCK) (_SOCK < 0)
#define SOCKET_ERROR_NUMBER errno
#define INVALID_SOCKET -1
#define THREAD pthread_t
#endif

static char *SUT_portNames[] = {"UDP1", "UDP2", "UDP3"};
static unsigned short SUT_portNumbers[] = {5061, 5062, 5063};

static SOCKET SUT_sockets[] = {INVALID_SOCKET, INVALID_SOCKET, INVALID_SOCKET};
static THREAD SUT_thread[UDP3+1];

static RTDS_TTCN_SystemInterfacePort SUT_systemPorts[UDP3+1];
RTDS_TTCN_SystemInterfacePort *RTDS_TTCN_systemInterfacePorts = NULL;

static void d_sip_request_line(sip_t *sip, Request *request)
  {
  switch (sip->sip_request->rq_method)
    {
    case sip_method_ack: request->requestLine.method = TTCN_constant_ACK_E; break;
    case sip_method_bye: request->requestLine.method = TTCN_constant_BYE_E; break;
    case sip_method_cancel: request->requestLine.method = TTCN_constant_CANCEL_E; break;
    case sip_method_invite: request->requestLine.method = TTCN_constant_INVITE_E; break;
    case sip_method_options: request->requestLine.method = TTCN_constant_OPTIONS_E; break;
    case sip_method_register: request->requestLine.method = TTCN_constant_REGISTER_E; break;
    case sip_method_prack: request->requestLine.method = TTCN_constant_PRACK_E; break;
    }
  strcpy(request->requestLine.requestUri.scheme, sip->sip_request->rq_url->url_scheme);
  request->requestLine.requestUri.userInfo.userOrTelephoneSubscriberPresent = (RTDS_BOOLEAN) (sip->sip_request->rq_url->url_user != NULL);
  if (request->requestLine.requestUri.userInfo.userOrTelephoneSubscriberPresent == TRUE)
    {
    strcpy(request->requestLine.requestUri.userInfo.userOrTelephoneSubscriber, sip->sip_request->rq_url->url_user);
    }
  request->requestLine.requestUri.userInfo.passwordPresent = (RTDS_BOOLEAN) (sip->sip_request->rq_url->url_password != NULL);
  if (request->requestLine.requestUri.userInfo.passwordPresent == TRUE)
    {
    strcpy(request->requestLine.requestUri.userInfo.password, sip->sip_request->rq_url->url_password);
    }
  request->requestLine.requestUri.userInfoPresent = (RTDS_BOOLEAN) (request->requestLine.requestUri.userInfo.userOrTelephoneSubscriberPresent ||
                                                request->requestLine.requestUri.userInfo.passwordPresent == TRUE);
  request->requestLine.requestUri.hostPort.hostPresent = (RTDS_BOOLEAN) (sip->sip_request->rq_url->url_host != NULL);
  if (request->requestLine.requestUri.hostPort.hostPresent == TRUE)
    {
    strcpy(request->requestLine.requestUri.hostPort.host, sip->sip_request->rq_url->url_host);
    }
  request->requestLine.requestUri.hostPort.portFieldPresent = (RTDS_BOOLEAN) (sip->sip_request->rq_url->url_port != NULL);
  if (request->requestLine.requestUri.hostPort.portFieldPresent == TRUE)
    {
    request->requestLine.requestUri.hostPort.portField = atoi(sip->sip_request->rq_url->url_port);
    }
  request->requestLine.requestUri.urlParametersPresent = (RTDS_BOOLEAN) (sip->sip_request->rq_url->url_params != NULL);
  if (request->requestLine.requestUri.urlParametersPresent == TRUE)
    {
    RTDS_SemicolonParam_List_init(&(request->requestLine.requestUri.urlParameters), TRUE);
    int i = 0;
    char *param = strtok((char *) sip->sip_request->rq_url->url_params, ";");
    while (param != NULL)
      {
      RTDS_SemicolonParam_List_add_index(&(request->requestLine.requestUri.urlParameters), i, NULL);
      char id[256], value[256];
      int p = sscanf(param, "%[^=]=%[^=]", id, value);
      ((GenericParam*)request->requestLine.requestUri.urlParameters.elements)[i].idPresent = (RTDS_BOOLEAN) (p > 0);
      if (((GenericParam*)request->requestLine.requestUri.urlParameters.elements)[i].idPresent == TRUE)
        {
        strcpy(((GenericParam*)request->requestLine.requestUri.urlParameters.elements)[i].id, id);
        }
      ((GenericParam*)request->requestLine.requestUri.urlParameters.elements)[i].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
      if (((GenericParam*)request->requestLine.requestUri.urlParameters.elements)[i].paramValuePresent == TRUE)
        {
        strcpy(((GenericParam*)request->requestLine.requestUri.urlParameters.elements)[i].paramValue, value);
        }
      ++i;
      param = strtok(NULL, ";");
      }
    }
  request->requestLine.requestUri.headersPresent = (RTDS_BOOLEAN) (sip->sip_request->rq_url->url_headers != NULL);
  if (request->requestLine.requestUri.headersPresent == TRUE)
    {
    RTDS_AmpersandParam_List_init(&(request->requestLine.requestUri.headers), TRUE);
    int i = 0;
    char *param = strtok((char *) sip->sip_request->rq_url->url_headers, "&");
    while (param != NULL)
      {
      RTDS_AmpersandParam_List_add_index(&(request->requestLine.requestUri.headers), i, NULL);
      char id[256], value[256];
      int p = sscanf(param, "%[^=]=%[^=]", id, value);
      ((GenericParam*)request->requestLine.requestUri.headers.elements)[i].idPresent = (RTDS_BOOLEAN) (p > 0);
      if (((GenericParam*)request->requestLine.requestUri.headers.elements)[i].idPresent == TRUE)
        {
        strcpy(((GenericParam*)request->requestLine.requestUri.headers.elements)[i].id, id);
        }
      ((GenericParam*)request->requestLine.requestUri.headers.elements)[i].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
      if (((GenericParam*)request->requestLine.requestUri.headers.elements)[i].paramValuePresent == TRUE)
        {
        strcpy(((GenericParam*)request->requestLine.requestUri.headers.elements)[i].paramValue, value);
        }
      ++i;
      param = strtok(NULL, "&");
      }
    }
  strcpy(request->requestLine.sipVersion, sip->sip_request->rq_version);
  }

static void e_sip_request_line(char *buffer, Request *request)
  {
  memset(buffer, 0, BUFFER_SIZE);
  switch (request->requestLine.method)
    {
    case TTCN_constant_ACK_E: strcpy(buffer, "ACK "); break;
    case TTCN_constant_BYE_E: strcpy(buffer, "BYE "); break;
    case TTCN_constant_CANCEL_E: strcpy(buffer, "CANCEL "); break;
    case TTCN_constant_INVITE_E: strcpy(buffer, "INVITE "); break;
    case TTCN_constant_OPTIONS_E: strcpy(buffer, "OPTIONS "); break;
    case TTCN_constant_REGISTER_E: strcpy(buffer, "REGISTER "); break;
    case TTCN_constant_PRACK_E: strcpy(buffer, "PRACK "); break;
    }
  if (!strcmp(request->requestLine.requestUri.scheme, "*"))
    {
    strcat(buffer, "sip:");
    }
  else
    {
    strcat(buffer, request->requestLine.requestUri.scheme);
    strcat(buffer, ":");
    }
  if (request->requestLine.requestUri.userInfo.userOrTelephoneSubscriberPresent == TRUE)
    {
    strcat(buffer, request->requestLine.requestUri.userInfo.userOrTelephoneSubscriber);
    if (request->requestLine.requestUri.userInfo.passwordPresent == TRUE)
      {
      strcat(buffer, ":");
      strcat(buffer, request->requestLine.requestUri.userInfo.password);
      }
    strcat(buffer, "@");
    }
  if (request->requestLine.requestUri.hostPort.hostPresent == TRUE)
    {
    strcat(buffer, request->requestLine.requestUri.hostPort.host);
    if (request->requestLine.requestUri.hostPort.portFieldPresent == TRUE)
      {
      strcat(buffer, ":");
      char port[8];
      sprintf(port, "%d", request->requestLine.requestUri.hostPort.portField);
      strcat(buffer, port);
      }
    }
  int i;
  if (request->requestLine.requestUri.urlParametersPresent == TRUE)
    {
    for (i=0; i<request->requestLine.requestUri.urlParameters.length; ++i)
      {
      if (((GenericParam*)request->requestLine.requestUri.urlParameters.elements)[i].idPresent == TRUE)
        {
        strcat(buffer, ";");
        strcat(buffer, ((GenericParam*)request->requestLine.requestUri.urlParameters.elements)[i].id);
        if (((GenericParam*)request->requestLine.requestUri.urlParameters.elements)[i].paramValuePresent == TRUE)
          {
          strcat(buffer, "=");
          strcat(buffer, ((GenericParam*)request->requestLine.requestUri.urlParameters.elements)[i].paramValue);
          }
        }
      }
    }
  if (request->requestLine.requestUri.headersPresent == TRUE)
    {
    for(i=0; i<request->requestLine.requestUri.headers.length; ++i)
      {
      if (((GenericParam*)request->requestLine.requestUri.headers.elements)[i].idPresent == TRUE)
        {
        if (i == 0)
          {
          strcat(buffer, "?");
          }
        else
          {
          strcat(buffer, "&");
          }
        strcat(buffer, ((GenericParam*)request->requestLine.requestUri.headers.elements)[i].id);
        if (((GenericParam*)request->requestLine.requestUri.headers.elements)[i].paramValuePresent == TRUE)
          {
          strcat(buffer, "=");
          strcat(buffer, ((GenericParam*)request->requestLine.requestUri.headers.elements)[i].paramValue);
          }
        }
      }
    }
  strcat(buffer, " ");
  strcat(buffer, request->requestLine.sipVersion);
  strcat(buffer, "\r\n");
  }

static void d_sip_status_line(sip_t *sip, Response *response)
  {
  if (sip->sip_status->st_version != NULL)
    {
    strcpy(response->statusLine.sipVersion, sip->sip_status->st_version);
    }
  response->statusLine.statusCode = sip->sip_status->st_status;
  if (sip->sip_status->st_phrase != NULL)
    {
    strcpy(response->statusLine.reasonPhrase, sip->sip_status->st_phrase);
    }
  }

static void e_sip_status_line(char *buffer, Response *response)
  {
  memset(buffer, 0, BUFFER_SIZE);
  strcpy(buffer, response->statusLine.sipVersion);
  strcat(buffer, " ");
  char code[8];
  sprintf(code, "%d", response->statusLine.statusCode);
  strcat(buffer, code);
  strcat(buffer, " ");
  strcat(buffer, response->statusLine.reasonPhrase);
  strcat(buffer, "\r\n");
  }
  
static void d_sip_accept(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_accept_t *header = sip->sip_accept;
  msgHeader->acceptPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->acceptPresent == TRUE)
    {
    msgHeader->accept.fieldName = TTCN_constant_ACCEPT_E;
    RTDS_AcceptBody_List_init(&(msgHeader->accept.acceptArgs), TRUE);
    int i = 0;
    while (header != NULL)
      {
      if (header->ac_type != NULL)
        {
        RTDS_AcceptBody_List_add_index(&(msgHeader->accept.acceptArgs), i, NULL);
        msgHeader->accept.acceptArgsPresent = TRUE;
        strcpy(((AcceptBody *)msgHeader->accept.acceptArgs.elements)[i].mediaRange, header->ac_type);
        ((AcceptBody *)msgHeader->accept.acceptArgs.elements)[i].acceptParamPresent = (RTDS_BOOLEAN) (header->ac_params != NULL);
        if (((AcceptBody *)msgHeader->accept.acceptArgs.elements)[i].acceptParamPresent == TRUE)
          {
          RTDS_SemicolonParam_List_init(&(((AcceptBody *)msgHeader->accept.acceptArgs.elements)[i].acceptParam), TRUE);
          int j = 0;
          while (header->ac_params[j])
            {
            RTDS_SemicolonParam_List_add_index(&(((AcceptBody *)msgHeader->accept.acceptArgs.elements)[i].acceptParam), j, NULL);
            char id[256], value[256];
            int p = sscanf(header->ac_params[j], "%[^=]=%[^=]", id, value);
            ((GenericParam *)((AcceptBody *)msgHeader->accept.acceptArgs.elements)[i].acceptParam.elements)[j].idPresent = (RTDS_BOOLEAN) (p > 0);
            if (((GenericParam *)((AcceptBody *)msgHeader->accept.acceptArgs.elements)[i].acceptParam.elements)[j].idPresent == TRUE)
              {
              strcpy(((GenericParam *)((AcceptBody *)msgHeader->accept.acceptArgs.elements)[i].acceptParam.elements)[j].id, id);
              }
            ((GenericParam *)((AcceptBody *)msgHeader->accept.acceptArgs.elements)[i].acceptParam.elements)[j].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
            if (((GenericParam *)((AcceptBody *)msgHeader->accept.acceptArgs.elements)[i].acceptParam.elements)[j].paramValuePresent == TRUE)
              {
              strcpy(((GenericParam *)((AcceptBody *)msgHeader->accept.acceptArgs.elements)[i].acceptParam.elements)[j].paramValue, value);
              }
            ++j;
            }
          }
        ++i;
        }
      header = header->ac_next;
      }
    }
  }

static void e_sip_accept(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->acceptPresent == TRUE)
    {
    strcat(buffer, "Accept: ");
    int i, j;
    for (i=0; i<msgHeader->accept.acceptArgs.length; ++i)
      {
      if (i > 0)
        {
        strcat(buffer, ",");
        }
      strcat(buffer, ((AcceptBody *)msgHeader->accept.acceptArgs.elements)[i].mediaRange);
      if (((AcceptBody *)msgHeader->accept.acceptArgs.elements)[i].acceptParamPresent == TRUE)
        {
        for (j=0; j<((AcceptBody *)msgHeader->accept.acceptArgs.elements)[i].acceptParam.length; ++j)
          {
          if (((GenericParam *)((AcceptBody *)msgHeader->accept.acceptArgs.elements)[i].acceptParam.elements)[j].idPresent == TRUE)
            {
            strcat(buffer, ";");
            strcat(buffer, ((GenericParam *)((AcceptBody *)msgHeader->accept.acceptArgs.elements)[i].acceptParam.elements)[j].id);
            if (((GenericParam *)((AcceptBody *)msgHeader->accept.acceptArgs.elements)[i].acceptParam.elements)[j].paramValuePresent == TRUE)
              {
              strcat(buffer, "=");
              strcat(buffer, ((GenericParam *)((AcceptBody *)msgHeader->accept.acceptArgs.elements)[i].acceptParam.elements)[j].paramValue);
              }
            }
          }
        }
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_accept_encoding(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_accept_encoding_t *header = sip->sip_accept_encoding;
  msgHeader->acceptEncodingPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->acceptEncodingPresent == TRUE)
    {
    msgHeader->acceptEncoding.fieldName = TTCN_constant_ACCEPT_ENCODING_E;
    RTDS_ContentCoding_List_init(&(msgHeader->acceptEncoding.contentCoding), TRUE);
    int i = 0;
    while (header != NULL)
      {
      if (header->aa_value)
        {
        RTDS_ContentCoding_List_add_index(&(msgHeader->acceptEncoding.contentCoding), i, NULL);
        msgHeader->acceptEncoding.contentCodingPresent = TRUE;
        strcpy(((ContentCoding *)msgHeader->acceptEncoding.contentCoding.elements)[i], header->aa_value);
        if (header->aa_params != NULL)
          {
          int j = 0;
          while (header->aa_params[j])
            {
            strcat(((ContentCoding *)msgHeader->acceptEncoding.contentCoding.elements)[i], ";");
            strcat(((ContentCoding *)msgHeader->acceptEncoding.contentCoding.elements)[i], header->aa_params[j]);
            ++j;
            }
          }
        ++i;
        }
      header = header->aa_next;
      }
    }
  }

static void e_sip_accept_encoding(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->acceptEncodingPresent == TRUE)
    {
    strcat(buffer, "Accept-Encoding: ");
    int i;
    for (i=0; i<msgHeader->acceptEncoding.contentCoding.length; ++i)
      {
      if (i > 0)
        {
        strcat(buffer, ",");
        }
      strcat(buffer, ((ContentCoding *)msgHeader->acceptEncoding.contentCoding.elements)[i]);
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_accept_language(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_accept_language_t *header = sip->sip_accept_language;
  msgHeader->acceptLanguagePresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->acceptLanguagePresent == TRUE)
    {
    msgHeader->acceptLanguage.fieldName = TTCN_constant_ACCEPT_LANGUAGE_E;
    RTDS_LanguageBody_List_init(&(msgHeader->acceptLanguage.languageBody), TRUE);
    int i = 0;
    while (header != NULL)
      {
      if (header->aa_value != NULL)
        {
        RTDS_LanguageBody_List_add_index(&(msgHeader->acceptLanguage.languageBody), i, NULL);
        msgHeader->acceptLanguage.languageBodyPresent = TRUE;
        strcpy(((LanguageBody *)msgHeader->acceptLanguage.languageBody.elements)[i].languageRange, header->aa_value);
        ((LanguageBody *)msgHeader->acceptLanguage.languageBody.elements)[i].acceptParamPresent = (RTDS_BOOLEAN) (header->aa_params != NULL);
        if (((LanguageBody *)msgHeader->acceptLanguage.languageBody.elements)[i].acceptParamPresent == TRUE)
          {
          RTDS_SemicolonParam_List_init(&(((LanguageBody *)msgHeader->acceptLanguage.languageBody.elements)[i].acceptParam), TRUE);
          int j = 0;
          while (header->aa_params[j])
            {
            RTDS_SemicolonParam_List_add_index(&(((LanguageBody *)msgHeader->acceptLanguage.languageBody.elements)[i].acceptParam), j, NULL);
            char id[256], value[256];
            int p = sscanf(header->aa_params[j], "%[^=]=%[^=]", id, value);
            ((GenericParam *)((LanguageBody *)msgHeader->acceptLanguage.languageBody.elements)[i].acceptParam.elements)[j].idPresent = (RTDS_BOOLEAN) (p > 0);
            if (((GenericParam *)((LanguageBody *)msgHeader->acceptLanguage.languageBody.elements)[i].acceptParam.elements)[j].idPresent == TRUE)
              {
              strcpy(((GenericParam *)((LanguageBody *)msgHeader->acceptLanguage.languageBody.elements)[i].acceptParam.elements)[j].id, id);
              }
            ((GenericParam *)((LanguageBody *)msgHeader->acceptLanguage.languageBody.elements)[i].acceptParam.elements)[j].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
            if (((GenericParam *)((LanguageBody *)msgHeader->acceptLanguage.languageBody.elements)[i].acceptParam.elements)[j].paramValuePresent == TRUE)
              {
              strcpy(((GenericParam *)((LanguageBody *)msgHeader->acceptLanguage.languageBody.elements)[i].acceptParam.elements)[j].paramValue, value);
              }
            ++j;
            }
          }
        ++i;
        }
      header = header->aa_next;
      }
    }
  }

static void e_sip_accept_language(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->acceptLanguagePresent == TRUE)
    {
    strcat(buffer, "Accept-Language: ");
    int i, j;
    for (i=0; i<msgHeader->acceptLanguage.languageBody.length; ++i)
      {
      if (i > 0)
        {
        strcat(buffer, ",");
        }
      strcat(buffer, ((LanguageBody *)msgHeader->acceptLanguage.languageBody.elements)[i].languageRange);
      if (((LanguageBody *)msgHeader->acceptLanguage.languageBody.elements)[i].acceptParamPresent == TRUE)
        {
        for (j=0; j<((LanguageBody *)msgHeader->acceptLanguage.languageBody.elements)[i].acceptParam.length; ++j)
          {
          if (((GenericParam *)((LanguageBody *)msgHeader->acceptLanguage.languageBody.elements)[i].acceptParam.elements)[j].idPresent == TRUE)
            {
            strcat(buffer, ";");
            strcat(buffer, ((GenericParam *)((LanguageBody *)msgHeader->acceptLanguage.languageBody.elements)[i].acceptParam.elements)[j].id);
            if (((GenericParam *)((LanguageBody *)msgHeader->acceptLanguage.languageBody.elements)[i].acceptParam.elements)[j].paramValuePresent == TRUE)
              {
              strcat(buffer, "=");
              strcat(buffer, ((GenericParam *)((LanguageBody *)msgHeader->acceptLanguage.languageBody.elements)[i].acceptParam.elements)[j].paramValue);
              }
            }
          }
        }
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_alert_info(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_alert_info_t *header = sip_alert_info(sip);
  msgHeader->alertInfoPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->alertInfoPresent == TRUE)
    {
    msgHeader->alertInfo.fieldName = TTCN_constant_ALERT_INFO_E;
    msgHeader->alertInfo.alertInfoBodyPresent = TRUE;
    RTDS_AlertInfoBody_List_init(&(msgHeader->alertInfo.alertInfoBody), TRUE);
    int i = 0;
    while (header != NULL)
      {
      RTDS_AlertInfoBody_List_add_index(&(msgHeader->alertInfo.alertInfoBody), i, NULL);
      strcpy(((AlertInfoBody *)msgHeader->alertInfo.alertInfoBody.elements)[i].url, url_as_string(msg_home((msg_t *) sip), header->ai_url));
      ((AlertInfoBody *)msgHeader->alertInfo.alertInfoBody.elements)[i].genericParamsPresent = (RTDS_BOOLEAN) (header->ai_params != NULL);
      if (((AlertInfoBody *)msgHeader->alertInfo.alertInfoBody.elements)[i].genericParamsPresent == TRUE)
        {
        RTDS_SemicolonParam_List_init(&(((AlertInfoBody *)msgHeader->alertInfo.alertInfoBody.elements)[i].genericParams), TRUE);
        int j = 0;
        while (header->ai_params[j])
          {
          RTDS_SemicolonParam_List_add_index(&(((AlertInfoBody *)msgHeader->alertInfo.alertInfoBody.elements)[i].genericParams), j, NULL);
          char id[256], value[256];
          int p = sscanf(header->ai_params[j], "%[^=]=%[^=]", id, value);
          ((GenericParam *)((AlertInfoBody *)msgHeader->alertInfo.alertInfoBody.elements)[i].genericParams.elements)[j].idPresent = (RTDS_BOOLEAN) (p > 0);
          if (((GenericParam *)((AlertInfoBody *)msgHeader->alertInfo.alertInfoBody.elements)[i].genericParams.elements)[j].idPresent == TRUE)
            {
            strcpy(((GenericParam *)((AlertInfoBody *)msgHeader->alertInfo.alertInfoBody.elements)[i].genericParams.elements)[j].id, id);
            }
          ((GenericParam *)((AlertInfoBody *)msgHeader->alertInfo.alertInfoBody.elements)[i].genericParams.elements)[j].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
          if (((GenericParam *)((AlertInfoBody *)msgHeader->alertInfo.alertInfoBody.elements)[i].genericParams.elements)[j].paramValuePresent == TRUE)
            {
            strcpy(((GenericParam *)((AlertInfoBody *)msgHeader->alertInfo.alertInfoBody.elements)[i].genericParams.elements)[j].paramValue, value);
            }
          ++j;
          }
        }
      ++i;
      header = header->ai_next;
      }
    }
  }

static void e_sip_alert_info(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->alertInfoPresent == TRUE)
    {
    strcat(buffer, "Alert-Info: ");
    int i, j;
    for (i=0; i<msgHeader->alertInfo.alertInfoBody.length; ++i)
      {
      if (i > 0)
        {
        strcat(buffer, ",");
        }
      strcat(buffer, "<");
      strcat(buffer, ((AlertInfoBody *)msgHeader->alertInfo.alertInfoBody.elements)[i].url);
      strcat(buffer, ">");
      if (((AlertInfoBody *)msgHeader->alertInfo.alertInfoBody.elements)[i].genericParamsPresent == TRUE)
        {
        for (j=0; j<((AlertInfoBody *)msgHeader->alertInfo.alertInfoBody.elements)[i].genericParams.length; ++j)
          {
          if (((GenericParam *)((AlertInfoBody *)msgHeader->alertInfo.alertInfoBody.elements)[i].genericParams.elements)[j].idPresent == TRUE)
            {
            strcat(buffer, ";");
            strcat(buffer, ((GenericParam *)((AlertInfoBody *)msgHeader->alertInfo.alertInfoBody.elements)[i].genericParams.elements)[j].id);
            if (((GenericParam *)((AlertInfoBody *)msgHeader->alertInfo.alertInfoBody.elements)[i].genericParams.elements)[j].paramValuePresent == TRUE)
              {
              strcat(buffer, ";");
              strcat(buffer, ((GenericParam *)((AlertInfoBody *)msgHeader->alertInfo.alertInfoBody.elements)[i].genericParams.elements)[j].paramValue);
              }
            }
          }
        }
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_allow(sip_t *sip, MessageHeader *msgHeader)
  {
  msg_list_t *header = (msg_list_t *) sip->sip_allow;
  msgHeader->allowPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->allowPresent == TRUE)
    {
    msgHeader->allow.fieldName = TTCN_constant_ALLOW_E;
    RTDS_Method_List_init(&(msgHeader->allow.methods), TRUE);
    int i = 0;
    while (header != NULL)
      {
      if (header->k_items != NULL)
        {
        msgHeader->allow.methodsPresent = TRUE;
        int j = 0;
        while (header->k_items[j])
          {
          RTDS_Method_List_add_index(&(msgHeader->allow.methods), i, NULL);
          strcpy(((RTDS_String *)msgHeader->allow.methods.elements)[i], header->k_items[j]);
          ++j;
          ++i;
          }
        }
      header = header->k_next;
      }
    }
  }

static void e_sip_allow(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->allowPresent == TRUE)
    {
    strcat(buffer, "Allow: ");
    int i;
    for (i=0; i<msgHeader->allow.methods.length; ++i)
      {
      if (i > 0)
        {
        strcat(buffer, ",");
        }
      strcat(buffer, ((RTDS_String *)msgHeader->allow.methods.elements)[i]);
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_authentication_info(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_authentication_info_t *header = sip->sip_authentication_info;
  msgHeader->authenticationInfoPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->authenticationInfoPresent == TRUE)
    {
    msgHeader->authenticationInfo.fieldName = TTCN_constant_AUTHENTICATION_INFO_E;
    RTDS_CommaParam_List_init(&(msgHeader->authenticationInfo.ainfo), TRUE);
    if (header->ai_params != NULL)
      {
      int i = 0;
      while (header->ai_params[i])
        {
        RTDS_CommaParam_List_add_index(&(msgHeader->authenticationInfo.ainfo), i, NULL);
        char id[256], value[256];
        int p = sscanf(header->ai_params[i], "%[^=]=%[^=]", id, value);
        ((GenericParam *)msgHeader->authenticationInfo.ainfo.elements)[i].idPresent = (RTDS_BOOLEAN) (p > 0);
        if (((GenericParam *)msgHeader->authenticationInfo.ainfo.elements)[i].idPresent == TRUE)
          {
          strcpy(((GenericParam *)msgHeader->authenticationInfo.ainfo.elements)[i].id, id);
          }
        ((GenericParam *)msgHeader->authenticationInfo.ainfo.elements)[i].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
        if (((GenericParam *)msgHeader->authenticationInfo.ainfo.elements)[i].paramValuePresent == TRUE)
          {
          strcpy(((GenericParam *)msgHeader->authenticationInfo.ainfo.elements)[i].paramValue, value);
          }
        ++i;
        }
      }
    }
  }

static void e_sip_authentication_info(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->authenticationInfoPresent == TRUE)
    {
    strcat(buffer, "Authentication-Info: ");
    int i;
    for (i=0; i<msgHeader->authenticationInfo.ainfo.length; ++i)
      {
      if (((GenericParam *)msgHeader->authenticationInfo.ainfo.elements)[i].idPresent == TRUE)
        {
        if (i > 0)
          {
          strcat(buffer, ",");
          }
        strcat(buffer, ((GenericParam *)msgHeader->authenticationInfo.ainfo.elements)[i].id);
        if (((GenericParam *)msgHeader->authenticationInfo.ainfo.elements)[i].paramValuePresent == TRUE)
          {
          strcat(buffer, "=");
          strcpy(buffer, ((GenericParam *)msgHeader->authenticationInfo.ainfo.elements)[i].paramValue);
          }
        }
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_authorization(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_authorization_t *header = sip->sip_authorization;
  msgHeader->authorizationPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->authorizationPresent == TRUE)
    {
    msgHeader->authorization.fieldName = TTCN_constant_AUTHORIZATION_E;
    msgHeader->authorization.bodyPresent = (RTDS_BOOLEAN) (header->au_scheme != NULL);
    if (msgHeader->authorization.bodyPresent == TRUE)
      {
      strcpy(msgHeader->authorization.body, header->au_scheme);
      if (header->au_params != NULL)
        {
        int i = 0;
        while (header->au_params[i])
          {
          if (i == 0)
            {
            strcat(msgHeader->authorization.body, " ");
            }
          else
            {
            strcat(msgHeader->authorization.body, ",");
            }
          strcat(msgHeader->authorization.body, header->au_params[i]);
          ++i;
          }
        }
      }
    }
  }

static void e_sip_authorization(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->authorizationPresent == TRUE)
    {
    strcat(buffer, "Authorization: ");
    if (msgHeader->authorization.bodyPresent == TRUE)
      {
      strcat(buffer, msgHeader->authorization.body);
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_call_id(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_call_id_t *header = sip->sip_call_id;
  msgHeader->callIdPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->callIdPresent == TRUE)
    {
    msgHeader->callId.fieldName = TTCN_constant_CALL_ID_E;
    msgHeader->callId.fieldNamePresent = TRUE;
    msgHeader->callId.callidPresent = (RTDS_BOOLEAN) (header->i_id != NULL);
    if (msgHeader->callId.callidPresent == TRUE)
      {
      strcpy(msgHeader->callId.callid, header->i_id);
      }
    }
  }

static void e_sip_call_id(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->callIdPresent == TRUE)
    {
    strcat(buffer, "Call-ID: ");
    strcat(buffer, msgHeader->callId.callid);
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_call_info(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_call_info_t *header = sip->sip_call_info;
  msgHeader->callInfoPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->callInfoPresent == TRUE)
    {
    msgHeader->callInfo.fieldName = TTCN_constant_CALL_INFO_E;
    msgHeader->callInfo.callInfoBodyPresent = TRUE;
    RTDS_CallInfoBody_List_init(&(msgHeader->callInfo.callInfoBody), TRUE);
    int i = 0;
    while (header != NULL)
      {
      RTDS_CallInfoBody_List_add_index(&(msgHeader->callInfo.callInfoBody), i, NULL);
      if (header->ci_url != NULL)
        {
        url_e(((CallInfoBody *)msgHeader->callInfo.callInfoBody.elements)[i].url, sizeof(RTDS_String), header->ci_url);
        }
      ((CallInfoBody *)msgHeader->callInfo.callInfoBody.elements)[i].infoParamsPresent = (RTDS_BOOLEAN) (header->ci_params != NULL);
      if (((CallInfoBody *)msgHeader->callInfo.callInfoBody.elements)[i].infoParamsPresent == TRUE)
        {
        RTDS_SemicolonParam_List_init(&(((CallInfoBody *)msgHeader->callInfo.callInfoBody.elements)[i].infoParams), TRUE);
        int j = 0;
        while (header->ci_params[j])
          {
          RTDS_SemicolonParam_List_add_index(&(((CallInfoBody *)msgHeader->callInfo.callInfoBody.elements)[i].infoParams), j, NULL);
          char id[256], value[256];
          int p = sscanf(header->ci_params[j], "%[^=]=%[^=]", id, value);
          ((GenericParam *)((CallInfoBody *)msgHeader->callInfo.callInfoBody.elements)[i].infoParams.elements)[j].idPresent = (RTDS_BOOLEAN) (p > 0);
          if (((GenericParam *)((CallInfoBody *)msgHeader->callInfo.callInfoBody.elements)[i].infoParams.elements)[j].idPresent == TRUE)
            {
            strcpy(((GenericParam *)((CallInfoBody *)msgHeader->callInfo.callInfoBody.elements)[i].infoParams.elements)[j].id, id);
            }
          ((GenericParam *)((CallInfoBody *)msgHeader->callInfo.callInfoBody.elements)[i].infoParams.elements)[j].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
          if (((GenericParam *)((CallInfoBody *)msgHeader->callInfo.callInfoBody.elements)[i].infoParams.elements)[j].paramValuePresent == TRUE)
            {
            strcpy(((GenericParam *)((CallInfoBody *)msgHeader->callInfo.callInfoBody.elements)[i].infoParams.elements)[j].paramValue, value);
            }
          ++j;
          }
        }
      ++i;
      header = header->ci_next;
      }
    }
  }

static void e_sip_call_info(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->callInfoPresent == TRUE)
    {
    strcat(buffer, "Call-Info: ");
    int i, j;
    for (i=0; i<msgHeader->alertInfo.alertInfoBody.length; ++i)
      {
      if (i > 0)
        {
        strcat(buffer, ",");
        }
      strcat(buffer, "<");
      strcat(buffer, ((CallInfoBody *)msgHeader->callInfo.callInfoBody.elements)[i].url);
      strcat(buffer, ">");
      if (((CallInfoBody *)msgHeader->callInfo.callInfoBody.elements)[i].infoParamsPresent == TRUE)
        {
        for (j=0; j<((CallInfoBody *)msgHeader->callInfo.callInfoBody.elements)[i].infoParams.length; ++j)
          {
          if (((GenericParam *)((CallInfoBody *)msgHeader->callInfo.callInfoBody.elements)[i].infoParams.elements)[j].idPresent == TRUE)
            {
            strcat(buffer, ";");
            strcat(buffer, ((GenericParam *)((CallInfoBody *)msgHeader->callInfo.callInfoBody.elements)[i].infoParams.elements)[j].id);
            if (((GenericParam *)((CallInfoBody *)msgHeader->callInfo.callInfoBody.elements)[i].infoParams.elements)[j].paramValuePresent == TRUE)
              {
              strcat(buffer, "=");
              strcat(buffer, ((GenericParam *)((CallInfoBody *)msgHeader->callInfo.callInfoBody.elements)[i].infoParams.elements)[j].paramValue);
              }
            }
          }
        }
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_contact(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_contact_t *header = sip->sip_contact;  
  msgHeader->contactPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->contactPresent == TRUE)
    {
    msgHeader->contact.fieldName = TTCN_constant_CONTACT_E;
    RTDS_ContactAddress_List_init(&(msgHeader->contact.contactBody.__value.contactAddresses), TRUE);
    int i = 0;
    while (header != NULL)
      {
      if (header->m_url->url_scheme != NULL && !strcmp(header->m_url->url_scheme, "*"))
        {
        msgHeader->contact.contactBody.present = TTCN_constant_ContactBody_wildcard;
        strcpy(msgHeader->contact.contactBody.__value.wildcard, "*");
        return;
        }
      else
        {
        msgHeader->contact.contactBody.present = TTCN_constant_ContactBody_contactAddresses;
        RTDS_ContactAddress_List_add_index(&(msgHeader->contact.contactBody.__value.contactAddresses), i, NULL);
        SipUrl *url;
        if (header->m_display != NULL)
          {
          ((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].addressField.present = TTCN_constant_Addr_Union_nameAddr;
          ((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].addressField.__value.nameAddr.displayNamePresent = TRUE;
          strcpy(((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].addressField.__value.nameAddr.displayName, header->m_display);
          url = &((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].addressField.__value.nameAddr.addrSpec;
          }
        else
          {
          ((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].addressField.present = TTCN_constant_Addr_Union_addrSpecUnion;
          url = &((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].addressField.__value.addrSpecUnion;
          }
        if (header->m_url->url_scheme != NULL)
          {
          strcpy(url->scheme, header->m_url->url_scheme);
          }
        url->userInfo.userOrTelephoneSubscriberPresent = (RTDS_BOOLEAN) (header->m_url->url_user != NULL);
        if (url->userInfo.userOrTelephoneSubscriberPresent == TRUE)
          {
          strcpy(url->userInfo.userOrTelephoneSubscriber, header->m_url->url_user);
          }
        url->userInfo.passwordPresent = (RTDS_BOOLEAN) (header->m_url->url_password != NULL);
        if (url->userInfo.passwordPresent == TRUE)
          {
          strcpy(url->userInfo.password, header->m_url->url_password);
          }
        url->userInfoPresent = (RTDS_BOOLEAN) (url->userInfo.userOrTelephoneSubscriberPresent || url->userInfo.passwordPresent == TRUE);
        url->hostPort.hostPresent = (RTDS_BOOLEAN) (header->m_url->url_host != NULL);
        if (url->hostPort.hostPresent == TRUE)
          {
          strcpy(url->hostPort.host, header->m_url->url_host);
          }
        url->hostPort.portFieldPresent = (RTDS_BOOLEAN) (header->m_url->url_port != NULL);
        if (url->hostPort.portFieldPresent == TRUE)
          {
          url->hostPort.portField = atoi(header->m_url->url_port);
          }
        url->urlParametersPresent = (RTDS_BOOLEAN) (header->m_url->url_params != NULL);
        if (url->urlParametersPresent == TRUE)
          {
          RTDS_SemicolonParam_List_init(&(url->urlParameters), TRUE);
          int j = 0;
          char *param = strtok((char *) header->m_url->url_params, ";");
          while (param != NULL)
            {
            RTDS_SemicolonParam_List_add_index(&(url->urlParameters), j, NULL);
            char id[256], value[256];
            int p = sscanf(param, "%[^=]=%[^=]", id, value);
            ((GenericParam *)url->urlParameters.elements)[j].idPresent = (RTDS_BOOLEAN) (p > 0);
            if (((GenericParam *)url->urlParameters.elements)[j].idPresent == TRUE)
              {
              strcpy(((GenericParam *)url->urlParameters.elements)[j].id, id);
              }
            ((GenericParam *)url->urlParameters.elements)[j].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
            if (((GenericParam *)url->urlParameters.elements)[j].paramValuePresent == TRUE)
              {
              strcpy(((GenericParam *)url->urlParameters.elements)[j].paramValue, value);
              }
            ++j;
            param = strtok(NULL, ";");
            }
          }
        url->headersPresent = (RTDS_BOOLEAN) (header->m_url->url_headers != NULL);
        if (url->headersPresent == TRUE)
          {
          RTDS_AmpersandParam_List_init(&(url->headers), TRUE);
          int j = 0;
          char *param = strtok((char *) header->m_url->url_headers, "&");
          while (param != NULL)
            {
            RTDS_AmpersandParam_List_add_index(&(url->headers), j, NULL);
            char id[256], value[256];
            int p = sscanf(param, "%[^=]=%[^=]", id, value);
            ((GenericParam *)url->headers.elements)[j].idPresent = (RTDS_BOOLEAN) (p > 0);
            if (((GenericParam *)url->headers.elements)[j].idPresent == TRUE)
              {
              strcpy(((GenericParam *)url->headers.elements)[j].id, id);
              }
            ((GenericParam *)url->headers.elements)[j].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
            if (((GenericParam *)url->headers.elements)[j].paramValuePresent == TRUE)
              {
              strcpy(((GenericParam *)url->headers.elements)[j].paramValue, value);
              }
            ++j;
            param = strtok(NULL, "&");
            }
          }
        ((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].contactParamsPresent = (RTDS_BOOLEAN) (header->m_params != NULL);
        if (((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].contactParamsPresent == TRUE)
          {
          RTDS_SemicolonParam_List_init(&(((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].contactParams), TRUE);
          int j = 0;
          while (header->m_params[j])
            {
            RTDS_SemicolonParam_List_add_index(&(((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].contactParams), j, NULL);
            char id[256], value[256];
            int p = sscanf(header->m_params[j], "%[^=]=%[^=]", id, value);
            ((GenericParam *)((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].contactParams.elements)[j].idPresent = (RTDS_BOOLEAN) (p > 0);
            if (((GenericParam *)((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].contactParams.elements)[j].idPresent == TRUE)
              {
              strcpy(((GenericParam *)((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].contactParams.elements)[j].id, id);
              }
            ((GenericParam *)((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].contactParams.elements)[j].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
            if (((GenericParam *)((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].contactParams.elements)[j].paramValuePresent == TRUE)
              {
              strcpy(((GenericParam *)((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].contactParams.elements)[j].paramValue, value);
              }
            ++j;
            }
          }
        ++i;
        }
      header = header->m_next;
      }
    }
  }

static void e_sip_contact(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->contactPresent == TRUE)
    {
    strcat(buffer, "Contact: ");
    if (msgHeader->contact.contactBody.present == TTCN_constant_ContactBody_wildcard)
      {
      strcat(buffer, "*");
      strcat(buffer, "\r\n");
      return;
      }
    int i, j;
    for (i=0; i<msgHeader->contact.contactBody.__value.contactAddresses.length; ++i)
      {
      if (i > 0)
        {
        strcat(buffer, ",");
        }
      SipUrl *url;
      if (((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].addressField.present == TTCN_constant_Addr_Union_nameAddr)
        {
        strcat(buffer, ((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].addressField.__value.nameAddr.displayName);
        strcat(buffer, "<");
        url = &((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].addressField.__value.nameAddr.addrSpec;
        }
      else
        {
        url = &((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].addressField.__value.addrSpecUnion;
        }
      if (strcmp(url->scheme, ""))
        {
        strcat(buffer, url->scheme);
        strcat(buffer, ":");
        }
      if (url->userInfo.userOrTelephoneSubscriberPresent == TRUE)
        {
        strcat(buffer, url->userInfo.userOrTelephoneSubscriber);
        if (url->userInfo.passwordPresent == TRUE)
          {
          strcat(buffer, ":");
          strcat(buffer, url->userInfo.password);
          }
        strcat(buffer, "@");
        }
      if (url->hostPort.hostPresent == TRUE)
        {
        strcat(buffer, url->hostPort.host);
        if (url->hostPort.portFieldPresent == TRUE)
          {
          strcat(buffer, ":");
          char port[8];
          sprintf(port, "%d", url->hostPort.portField);
          strcat(buffer, port);
          }
        }
      if (url->urlParametersPresent == TRUE)
        {
        for (j=0; j<url->urlParameters.length; ++j)
          {
          if (((GenericParam *)url->urlParameters.elements)[j].idPresent == TRUE)
            {
            strcat(buffer, ";");
            strcat(buffer, ((GenericParam *)url->urlParameters.elements)[j].id);
            if (((GenericParam *)url->urlParameters.elements)[j].paramValuePresent == TRUE)
              {
              strcat(buffer, "=");
              strcat(buffer, ((GenericParam *)url->urlParameters.elements)[j].paramValue);
              }
            }
          }
        }
      if (url->headersPresent == TRUE)
        {
        for(j=0; j<url->headers.length; ++j)
          {
          if (((GenericParam *)url->headers.elements)[j].idPresent == TRUE)
            {
            if (j == 0)
              {
              strcat(buffer, "?");
              }
            else
              {
              strcat(buffer, "&");
              }
            strcat(buffer, ((GenericParam *)url->headers.elements)[j].id);
            if (((GenericParam *)url->headers.elements)[j].paramValuePresent == TRUE)
              {
              strcat(buffer, "=");
              strcat(buffer, ((GenericParam *)url->headers.elements)[j].paramValue);
              }
            }
          }
        }
      if (((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].addressField.present == TTCN_constant_Addr_Union_nameAddr)
        {
        strcat(buffer, ">");
        }
      if (((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].contactParamsPresent == TRUE)
        {
        for (j=0; j<((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].contactParams.length; ++j)
          {
          if (((GenericParam *)((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].contactParams.elements)[j].idPresent == TRUE)
            {
            strcat(buffer, ";");
            strcat(buffer, ((GenericParam *)((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].contactParams.elements)[j].id);
            if (((GenericParam *)((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].contactParams.elements)[j].paramValuePresent == TRUE)
              {
              strcat(buffer, "=");
              strcat(buffer, ((GenericParam *)((ContactAddress *)msgHeader->contact.contactBody.__value.contactAddresses.elements)[i].contactParams.elements)[j].paramValue);
              }
            }
          }
        }
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_content_disposition(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_content_disposition_t *header = sip->sip_content_disposition;
  msgHeader->contentDispositionPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->contentDispositionPresent == TRUE)
    {
    msgHeader->contentDisposition.fieldName = TTCN_constant_CONTENT_DISPOSITION_E;
    strcpy(msgHeader->contentDisposition.dispositionType, header->cd_type);
    msgHeader->contentDisposition.dispositionParamsPresent = (RTDS_BOOLEAN) (header->cd_params != NULL);
    if (msgHeader->contentDisposition.dispositionParamsPresent == TRUE)
      {
      RTDS_SemicolonParam_List_init(&(msgHeader->contentDisposition.dispositionParams), TRUE);
      int i = 0;
      while (header->cd_params[i])
        {
        RTDS_SemicolonParam_List_add_index(&(msgHeader->contentDisposition.dispositionParams), i, NULL);
        char id[256], value[256];
        int p = sscanf(header->cd_params[i], "%[^=]=%[^=]", id, value);
        ((GenericParam *)msgHeader->contentDisposition.dispositionParams.elements)[i].idPresent = (RTDS_BOOLEAN) (p > 0);
        if (((GenericParam *)msgHeader->contentDisposition.dispositionParams.elements)[i].idPresent == TRUE)
          {
          strcpy(((GenericParam *)msgHeader->contentDisposition.dispositionParams.elements)[i].id, id);
          }
        ((GenericParam *)msgHeader->contentDisposition.dispositionParams.elements)[i].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
        if (((GenericParam *)msgHeader->contentDisposition.dispositionParams.elements)[i].paramValuePresent == TRUE)
          {
          strcpy(((GenericParam *)msgHeader->contentDisposition.dispositionParams.elements)[i].paramValue, value);
          }
        ++i;
        }
      }
    }
  }

static void e_sip_content_disposition(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->contentDispositionPresent == TRUE)
    {
    strcat(buffer, "Content-Disposition: ");
    strcat(buffer, msgHeader->contentDisposition.dispositionType);
    int i;
    if (msgHeader->contentDisposition.dispositionParamsPresent == TRUE)
      {
      for (i=0; i<msgHeader->contentDisposition.dispositionParams.length; ++i)
        {
        if (((GenericParam *)msgHeader->contentDisposition.dispositionParams.elements)[i].idPresent == TRUE)
          {
          strcat(buffer, ";");
          strcat(buffer, ((GenericParam *)msgHeader->contentDisposition.dispositionParams.elements)[i].id);
          if (((GenericParam *)msgHeader->contentDisposition.dispositionParams.elements)[i].paramValuePresent == TRUE)
            {
            strcat(buffer, "=");
            strcat(buffer, ((GenericParam *)msgHeader->contentDisposition.dispositionParams.elements)[i].paramValue);
            }
          }
        }
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_content_encoding(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_content_encoding_t *header = sip->sip_content_encoding;
  msgHeader->contentEncodingPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->contentEncodingPresent == TRUE)
    {
    msgHeader->contentEncoding.fieldName = TTCN_constant_CONTENT_ENCODING_E;
    RTDS_ContentCoding_List_init(&(msgHeader->contentEncoding.contentCoding), TRUE);
    int i = 0;
    while (header != NULL)
      {
      if (header->k_items != NULL)
        {
        int j = 0;
        while (header->k_items[j])
          {
          RTDS_ContentCoding_List_add_index(&(msgHeader->contentEncoding.contentCoding), i, NULL);
          strcpy(((ContentCoding *)msgHeader->contentEncoding.contentCoding.elements)[i], header->k_items[j]);
          ++j;
          ++i;
          }
        }
      header = header->k_next;
      }
    }
  }

static void e_sip_content_encoding(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->contentEncodingPresent == TRUE)
    {
    strcat(buffer, "Content-Encoding: ");
    int i;
    for (i=0; i<msgHeader->contentEncoding.contentCoding.length; ++i)
      {
      if (i > 0)
        {
        strcat(buffer, ",");
        }
      strcat(buffer, ((ContentCoding *)msgHeader->contentEncoding.contentCoding.elements)[i]);
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_content_language(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_content_language_t *header = sip->sip_content_language;
  msgHeader->contentLanguagePresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->contentLanguagePresent == TRUE)
    {
    msgHeader->contentLanguage.fieldName = TTCN_constant_CONTENT_LANGUAGE_E;
    RTDS_LanguageTag_List_init(&(msgHeader->contentLanguage.languageTag), TRUE);
    int i = 0;
    while (header != NULL)
      {
      if (header->k_items != NULL)
        {
        int j = 0;
        while (header->k_items[j])
          {
          RTDS_LanguageTag_List_add_index(&(msgHeader->contentLanguage.languageTag), i, NULL);
          strcpy(((LanguageTag *)msgHeader->contentLanguage.languageTag.elements)[i], header->k_items[j]);
          ++j;
          ++i;
          }
        }
      header = header->k_next;
      }
    }
  }

static void e_sip_content_language(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->contentLanguagePresent == TRUE)
    {
    strcat(buffer, "Content-Language: ");
    int i;
    for (i=0; i<msgHeader->contentLanguage.languageTag.length; ++i)
      {
      if (i > 0)
        {
        strcat(buffer, ",");
        }
      strcat(buffer, ((LanguageTag *)msgHeader->contentLanguage.languageTag.elements)[i]);
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_content_length(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_content_length_t *header = sip->sip_content_length;
  msgHeader->contentLengthPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->contentLengthPresent == TRUE)
    {
    msgHeader->contentLength.fieldName = TTCN_constant_CONTENT_LENGTH_E;
    msgHeader->contentLength.len = (int) header->l_length;
    }
  }

static void e_sip_content_length(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->contentLengthPresent == TRUE)
    {
    strcat(buffer, "Content-Length: ");
    char len[32];
    sprintf(len, "%d", msgHeader->contentLength.len);
    strcat(buffer, len);
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_content_type(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_content_type_t *header = sip->sip_content_type;
  msgHeader->contentTypePresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->contentTypePresent == TRUE)
    {
    msgHeader->contentType.fieldName = TTCN_constant_CONTENT_TYPE_E;
    msgHeader->contentType.mediaTypePresent = (RTDS_BOOLEAN) (header->c_type != NULL);
    if (msgHeader->contentType.mediaTypePresent == TRUE)
      {
      strcpy(msgHeader->contentType.mediaType, header->c_type);
      if (header->c_params != NULL)
        {
        int i = 0;
        while (header->c_params[i])
          {
          strcat(msgHeader->contentType.mediaType, ";");
          strcat(msgHeader->contentType.mediaType, header->c_params[i]);
          ++i;
          }
        }
      }
    }
  }

static void e_sip_content_type(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->contentTypePresent == TRUE)
    {
    strcat(buffer, "Content-Type: ");
    if (msgHeader->contentType.mediaTypePresent == TRUE)
      {
      strcat(buffer, msgHeader->contentType.mediaType);
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_cseq(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_cseq_t *header = sip->sip_cseq;
  msgHeader->cSeqPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->cSeqPresent == TRUE)
    {
    msgHeader->cSeq.fieldName = TTCN_constant_CSEQ_E;
    msgHeader->cSeq.seqNumber = (int) header->cs_seq;
    if (header->cs_method_name != NULL)
      {
      strcpy(msgHeader->cSeq.method, header->cs_method_name);
      }
    }
  }

static void e_sip_cseq(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->cSeqPresent == TRUE)
    {
    strcat(buffer, "CSeq: ");
    char cSeq[BUFFER_SIZE];
    sprintf(cSeq, "%d %s", msgHeader->cSeq.seqNumber, msgHeader->cSeq.method);
    strcat(buffer, cSeq);
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_date(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_date_t *header = sip->sip_date;
  msgHeader->datePresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->datePresent == TRUE)
    {
    msgHeader->date.fieldName = TTCN_constant_DATE_E;
    sip_date_e(msgHeader->date.sipDate, sizeof(RTDS_String), (msg_header_t *) &(header->d_time), 0);
    }
  }

static void e_sip_date(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->datePresent == TRUE)
    {
    strcat(buffer, "Date: ");
    strcat(buffer, msgHeader->date.sipDate);
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_error_info(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_call_info_t *header = (sip_call_info_t *) sip->sip_error_info;
  msgHeader->errorInfoPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->errorInfoPresent == TRUE)
    {
    msgHeader->errorInfo.fieldName = TTCN_constant_ERROR_INFO_E;
    msgHeader->errorInfo.errorInfoPresent = TRUE;
    RTDS_ErrorInfoBody_List_init(&(msgHeader->errorInfo.errorInfo), TRUE);
    int i = 0;
    while (header != NULL)
      {
      RTDS_ErrorInfoBody_List_add_index(&(msgHeader->errorInfo.errorInfo), i, NULL);
      if (header->ci_url != NULL)
        {
        url_e(((ErrorInfoBody *)msgHeader->errorInfo.errorInfo.elements)[i].uri, sizeof(RTDS_String), header->ci_url);
        }
      ((ErrorInfoBody *)msgHeader->errorInfo.errorInfo.elements)[i].genericParamsPresent = (RTDS_BOOLEAN) (header->ci_params != NULL);
      if (((ErrorInfoBody *)msgHeader->errorInfo.errorInfo.elements)[i].genericParamsPresent == TRUE)
        {
        RTDS_SemicolonParam_List_init(&(((ErrorInfoBody *)msgHeader->errorInfo.errorInfo.elements)[i].genericParams), TRUE);
        int j = 0;
        while (header->ci_params[j])
          {
          RTDS_SemicolonParam_List_add_index(&(((ErrorInfoBody *)msgHeader->errorInfo.errorInfo.elements)[i].genericParams), j, NULL);
          char id[256], value[256];
          int p = sscanf(header->ci_params[j], "%[^=]=%[^=]", id, value);
          ((GenericParam *)((ErrorInfoBody *)msgHeader->errorInfo.errorInfo.elements)[i].genericParams.elements)[j].idPresent = (RTDS_BOOLEAN) (p > 0);
          if (((GenericParam *)((ErrorInfoBody *)msgHeader->errorInfo.errorInfo.elements)[i].genericParams.elements)[j].idPresent == TRUE)
            {
            strcpy(((GenericParam *)((ErrorInfoBody *)msgHeader->errorInfo.errorInfo.elements)[i].genericParams.elements)[j].id, id);
            }
          ((GenericParam *)((ErrorInfoBody *)msgHeader->errorInfo.errorInfo.elements)[i].genericParams.elements)[j].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
          if (((GenericParam *)((ErrorInfoBody *)msgHeader->errorInfo.errorInfo.elements)[i].genericParams.elements)[j].paramValuePresent == TRUE)
            {
            strcpy(((GenericParam *)((ErrorInfoBody *)msgHeader->errorInfo.errorInfo.elements)[i].genericParams.elements)[j].paramValue, value);
            }
          ++j;
          }
        }
      ++i;
      header = header->ci_next;
      }
    }
  }

static void e_sip_error_info(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->errorInfoPresent == TRUE)
    {
    strcat(buffer, "Error-Info: ");
    int i, j;
    for (i=0; i<msgHeader->errorInfo.errorInfo.length; ++i)
      {
      if (i > 0)
        {
        strcat(buffer, ",");
        }
      strcat(buffer, "<");
      strcat(buffer, ((ErrorInfoBody *)msgHeader->errorInfo.errorInfo.elements)[i].uri);
      strcat(buffer, ">");
      if (((ErrorInfoBody *)msgHeader->errorInfo.errorInfo.elements)[i].genericParamsPresent == TRUE)
        {
        for (j=0; j<((ErrorInfoBody *)msgHeader->errorInfo.errorInfo.elements)[i].genericParams.length; ++j)
          {
          if (((GenericParam *)((ErrorInfoBody *)msgHeader->errorInfo.errorInfo.elements)[i].genericParams.elements)[j].idPresent == TRUE)
            {
            strcat(buffer, ";");
            strcat(buffer, ((GenericParam *)((ErrorInfoBody *)msgHeader->errorInfo.errorInfo.elements)[i].genericParams.elements)[j].id);
            if (((GenericParam *)((ErrorInfoBody *)msgHeader->errorInfo.errorInfo.elements)[i].genericParams.elements)[j].paramValuePresent == TRUE)
              {
              strcat(buffer, "=");
              strcat(buffer, ((GenericParam *)((ErrorInfoBody *)msgHeader->errorInfo.errorInfo.elements)[i].genericParams.elements)[j].paramValue);
              }
            }
          }
        }
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_expires(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_expires_t *header = sip->sip_expires;
  msgHeader->expiresPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->expiresPresent == TRUE)
    {
    msgHeader->expires.fieldName = TTCN_constant_EXPIRES_E;
    sip_date_e(msgHeader->date.sipDate, sizeof(RTDS_String), (msg_header_t *) &(header->ex_delta), 0);
    }
  }

static void e_sip_expires(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->expiresPresent == TRUE)
    {
    strcat(buffer, "Expires: ");
    strcat(buffer, msgHeader->date.sipDate);
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_from(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_from_t *header = sip->sip_from;  
  msgHeader->fromFieldPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->fromFieldPresent == TRUE)
    {
    msgHeader->fromField.fieldName = TTCN_constant_FROM_E;
    SipUrl *url;
    if (header->a_display != NULL)
      {
      msgHeader->fromField.addressField.present = TTCN_constant_Addr_Union_nameAddr;
      msgHeader->fromField.addressField.__value.nameAddr.displayNamePresent = TRUE;
      strcpy(msgHeader->fromField.addressField.__value.nameAddr.displayName, header->a_display);
      url = &msgHeader->fromField.addressField.__value.nameAddr.addrSpec;
      }
    else
      {
      msgHeader->fromField.addressField.present = TTCN_constant_Addr_Union_addrSpecUnion;
      url = &msgHeader->fromField.addressField.__value.addrSpecUnion;
      }
    if (header->a_url->url_scheme != NULL)
      {
      strcpy(url->scheme, header->a_url->url_scheme);
      }
    url->userInfo.userOrTelephoneSubscriberPresent = (RTDS_BOOLEAN) (header->a_url->url_user != NULL);
    if (url->userInfo.userOrTelephoneSubscriberPresent == TRUE)
      {
      strcpy(url->userInfo.userOrTelephoneSubscriber, header->a_url->url_user);
      }
    url->userInfo.passwordPresent = (RTDS_BOOLEAN) (header->a_url->url_password != NULL);
    if (url->userInfo.passwordPresent == TRUE)
      {
      strcpy(url->userInfo.password, header->a_url->url_password);
      }
    url->userInfoPresent = (RTDS_BOOLEAN) (url->userInfo.userOrTelephoneSubscriberPresent == TRUE || url->userInfo.passwordPresent == TRUE);
    url->hostPort.hostPresent = (RTDS_BOOLEAN) (header->a_url->url_host != NULL);
    if (url->hostPort.hostPresent == TRUE)
      {
      strcpy(url->hostPort.host, header->a_url->url_host);
      }
    url->hostPort.portFieldPresent = (RTDS_BOOLEAN) (header->a_url->url_port != NULL);
    if (url->hostPort.portFieldPresent == TRUE)
      {
      url->hostPort.portField = atoi(header->a_url->url_port);
      }
    url->urlParametersPresent = (RTDS_BOOLEAN) (header->a_url->url_params != NULL);
    if (url->urlParametersPresent == TRUE)
      {
      RTDS_SemicolonParam_List_init(&(url->urlParameters), TRUE);
      int i = 0;
      char *param = strtok((char *) header->a_url->url_params, ";");
      while (param != NULL)
        {
        RTDS_SemicolonParam_List_add_index(&(url->urlParameters), i, NULL);
        char id[256], value[256];
        int p = sscanf(param, "%[^=]=%[^=]", id, value);
        ((GenericParam *)url->urlParameters.elements)[i].idPresent = (RTDS_BOOLEAN) (p > 0);
        if (((GenericParam *)url->urlParameters.elements)[i].idPresent == TRUE)
          {
          strcpy(((GenericParam *)url->urlParameters.elements)[i].id, id);
          }
        ((GenericParam *)url->urlParameters.elements)[i].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
        if (((GenericParam *)url->urlParameters.elements)[i].paramValuePresent == TRUE)
          {
          strcpy(((GenericParam *)url->urlParameters.elements)[i].paramValue, value);
          }
        ++i;
        param = strtok(NULL, ";");
        }
      }
    url->headersPresent = (RTDS_BOOLEAN) (header->a_url->url_headers != NULL);
    if (url->headersPresent == TRUE)
      {
      RTDS_AmpersandParam_List_init(&(url->headers), TRUE);
      int i = 0;
      char *param = strtok((char *) header->a_url->url_headers, "&");
      while (param != NULL)
        {
        RTDS_AmpersandParam_List_add_index(&(url->headers), i, NULL);
        char id[256], value[256];
        int p = sscanf(param, "%[^=]=%[^=]", id, value);
        ((GenericParam *)url->headers.elements)[i].idPresent = (RTDS_BOOLEAN) (p > 0);
        if (((GenericParam *)url->headers.elements)[i].idPresent == TRUE)
          {
          strcpy(((GenericParam *)url->headers.elements)[i].id, id);
          }
        ((GenericParam *)url->headers.elements)[i].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
        if (((GenericParam *)url->headers.elements)[i].paramValuePresent == TRUE)
          {
          strcpy(((GenericParam *)url->headers.elements)[i].paramValue, value);
          }
        ++i;
        param = strtok(NULL, "&");
        }
      }
    msgHeader->fromField.fromParamsPresent = (RTDS_BOOLEAN) (header->a_params != NULL);
    if (msgHeader->fromField.fromParamsPresent == TRUE)
      {
      RTDS_SemicolonParam_List_init(&(msgHeader->fromField.fromParams), TRUE);
      int i = 0;
      while (header->a_params[i])
        {
        RTDS_SemicolonParam_List_add_index(&(msgHeader->fromField.fromParams), i, NULL);
        char id[256], value[256];
        int p = sscanf(header->a_params[i], "%[^=]=%[^=]", id, value);
        ((GenericParam *)msgHeader->fromField.fromParams.elements)[i].idPresent = (RTDS_BOOLEAN) (p > 0);
        if (((GenericParam *)msgHeader->fromField.fromParams.elements)[i].idPresent == TRUE)
          {
          strcpy(((GenericParam *)msgHeader->fromField.fromParams.elements)[i].id, id);
          }
        ((GenericParam *)msgHeader->fromField.fromParams.elements)[i].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
        if (((GenericParam *)msgHeader->fromField.fromParams.elements)[i].paramValuePresent == TRUE)
          {
          strcpy(((GenericParam *)msgHeader->fromField.fromParams.elements)[i].paramValue, value);
          }
        ++i;
        }
      }
    }
  }

static void e_sip_from(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->fromFieldPresent == TRUE)
    {
    strcat(buffer, "From: ");
    int i;
    SipUrl *url;
    if (msgHeader->fromField.addressField.present == TTCN_constant_Addr_Union_nameAddr)
      {
      strcat(buffer, msgHeader->fromField.addressField.__value.nameAddr.displayName);
      strcat(buffer, "<");
      url = &msgHeader->fromField.addressField.__value.nameAddr.addrSpec;
      }
    else
      {
      url = &msgHeader->fromField.addressField.__value.addrSpecUnion;
      }
    if (strcmp(url->scheme, ""))
      {
      strcat(buffer, url->scheme);
      strcat(buffer, ":");
      }
    if (url->userInfo.userOrTelephoneSubscriberPresent == TRUE)
      {
      strcat(buffer, url->userInfo.userOrTelephoneSubscriber);
      if (url->userInfo.passwordPresent == TRUE)
        {
        strcat(buffer, ":");
        strcat(buffer, url->userInfo.password);
        }
      strcat(buffer, "@");
      }
    if (url->hostPort.hostPresent == TRUE)
      {
      strcat(buffer, url->hostPort.host);
      if (url->hostPort.portFieldPresent == TRUE)
        {
        strcat(buffer, ":");
        char port[8];
        sprintf(port, "%d", url->hostPort.portField);
        strcat(buffer, port);
        }
      }
    if (url->urlParametersPresent == TRUE)
      {
      for (i=0; i<url->urlParameters.length; ++i)
        {
        if (((GenericParam *)url->urlParameters.elements)[i].idPresent == TRUE)
          {
          strcat(buffer, ";");
          strcat(buffer, ((GenericParam *)url->urlParameters.elements)[i].id);
          if (((GenericParam *)url->urlParameters.elements)[i].paramValuePresent == TRUE)
            {
            strcat(buffer, "=");
            strcat(buffer, ((GenericParam *)url->urlParameters.elements)[i].paramValue);
            }
          }
        }
      }
    if (url->headersPresent == TRUE)
      {
      for(i=0; i<url->headers.length; ++i)
        {
        if (((GenericParam *)url->headers.elements)[i].idPresent == TRUE)
          {
          if (i == 0)
            {
            strcat(buffer, "?");
            }
          else
            {
            strcat(buffer, "&");
            }
          strcat(buffer, ((GenericParam *)url->headers.elements)[i].id);
          if (((GenericParam *)url->headers.elements)[i].paramValuePresent == TRUE)
            {
            strcat(buffer, "=");
            strcat(buffer, ((GenericParam *)url->headers.elements)[i].paramValue);
            }
          }
        }
      }
    if (msgHeader->fromField.addressField.present == TTCN_constant_Addr_Union_nameAddr)
      {
      strcat(buffer, ">");
      }
    if (msgHeader->fromField.fromParamsPresent == TRUE)
      {
      for (i=0; i<msgHeader->fromField.fromParams.length; ++i)
        {
        if (((GenericParam *)msgHeader->fromField.fromParams.elements)[i].idPresent == TRUE)
          {
          strcat(buffer, ";");
          strcat(buffer, ((GenericParam *)msgHeader->fromField.fromParams.elements)[i].id);
          if (((GenericParam *)msgHeader->fromField.fromParams.elements)[i].paramValuePresent == TRUE)
            {
            strcat(buffer, "=");
            strcat(buffer, ((GenericParam *)msgHeader->fromField.fromParams.elements)[i].paramValue);
            }
          }
        }
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_in_reply_to(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_in_reply_to_t *header = sip->sip_in_reply_to;
  msgHeader->inReplyToPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->inReplyToPresent == TRUE)
    {
    msgHeader->inReplyTo.fieldName = TTCN_constant_IN_REPLY_TO_E;
    RTDS_CallidString_List_init(&(msgHeader->inReplyTo.callids), TRUE);
    int i = 0;
    while (header != NULL)
      {
      if (header->k_items != NULL)
        {
        int j = 0;
        while (header->k_items[j])
          {
          RTDS_CallidString_List_add_index(&(msgHeader->inReplyTo.callids), i, NULL);
          strcpy(((CallidString *)msgHeader->inReplyTo.callids.elements)[i], header->k_items[j]);
          ++j;
          ++i;
          }
        }
      header = header->k_next;
      }
    }
  }

static void e_sip_in_reply_to(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->inReplyToPresent == TRUE)
    {
    strcat(buffer, "In-Reply-To: ");
    int i;
    for (i=0; i<msgHeader->inReplyTo.callids.length; ++i)
      {
      if (i > 0)
        {
        strcat(buffer, ",");
        }
      strcat(buffer, ((CallidString *)msgHeader->inReplyTo.callids.elements)[i]);
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_max_forwards(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_max_forwards_t *header = sip->sip_max_forwards;
  msgHeader->maxForwardsPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->maxForwardsPresent == TRUE)
    {
    msgHeader->maxForwards.fieldName = TTCN_constant_MAX_FORWARDS_E;
    msgHeader->maxForwards.forwards = (int) header->mf_count;
    }
  }

static void e_sip_max_forwards(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->maxForwardsPresent == TRUE)
    {
    strcat(buffer, "Max-Forwards: ");
    char forwards[32];
    sprintf(forwards, "%d", msgHeader->maxForwards.forwards);
    strcat(buffer, forwards);
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_mime_version(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_mime_version_t *header = sip->sip_mime_version;
  msgHeader->mimeVersionPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->mimeVersionPresent == TRUE)
    {
    msgHeader->mimeVersion.fieldName = TTCN_constant_MIME_VERSION_E;
    if (header->g_string != NULL)
      {
      sscanf(header->g_string, "%d.%d", &msgHeader->mimeVersion.majorNumber, &msgHeader->mimeVersion.minorNumber);
      }
    }
  }

static void e_sip_mime_version(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->mimeVersionPresent == TRUE)
    {
    strcat(buffer, "MIME-Version: ");
    char mimeVersion[64];
    sprintf(mimeVersion, "%d.%d", msgHeader->mimeVersion.majorNumber, msgHeader->mimeVersion.minorNumber);
    strcat(buffer, mimeVersion);
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_min_expires(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_min_expires_t *header = sip->sip_min_expires;
  msgHeader->minExpiresPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->minExpiresPresent == TRUE)
    {
    msgHeader->minExpires.fieldName = TTCN_constant_MIN_EXPIRES_E;
    sprintf(msgHeader->minExpires.deltaSec, "%d", (int) header->me_delta);
    }
  }

static void e_sip_min_expires(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->minExpiresPresent == TRUE)
    {
    strcat(buffer, "Min-Expires: ");
    strcat(buffer, msgHeader->minExpires.deltaSec);
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_organization(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_organization_t *header = sip->sip_organization;
  msgHeader->organizationPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->organizationPresent == TRUE)
    {
    msgHeader->organization.fieldName = TTCN_constant_ORGANIZATION_E;
    if (header->g_string != NULL)
      {
      strcpy(msgHeader->organization.organization, header->g_string);
      }
    }
  }

static void e_sip_organization(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->organizationPresent == TRUE)
    {
    strcat(buffer, "Organization: ");
    strcat(buffer, msgHeader->organization.organization);
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_priority(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_priority_t *header = sip->sip_priority;
  msgHeader->priorityPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->priorityPresent == TRUE)
    {
    msgHeader->priority.fieldName = TTCN_constant_PRIORITY_E;
    if (header->g_string != NULL)
      {
      strcpy(msgHeader->priority.priorityValue, header->g_string);
      }
    }
  }

static void e_sip_priority(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->priorityPresent == TRUE)
    {
    strcat(buffer, "Priority: ");
    strcat(buffer, msgHeader->priority.priorityValue);
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_proxy_authenticate(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_proxy_authenticate_t *header = sip->sip_proxy_authenticate;
  msgHeader->proxyAuthenticatePresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->proxyAuthenticatePresent == TRUE)
    {
    msgHeader->proxyAuthenticate.fieldName = TTCN_constant_PROXY_AUTHENTICATE_E;
    CommaParam_List *params;
    if (header->au_scheme != NULL)
      {
      if (strncmp(header->au_scheme, "Digest", 6) == 0)
        {
        msgHeader->proxyAuthenticate.challenge.present = TTCN_constant_Challenge_digestCln;
        params = &msgHeader->proxyAuthenticate.challenge.__value.digestCln;
        }
      else
        {
        msgHeader->proxyAuthenticate.challenge.present = TTCN_constant_Challenge_otherChallenge;
        strcpy(msgHeader->proxyAuthenticate.challenge.__value.otherChallenge.authScheme, header->au_scheme);
        params = &msgHeader->proxyAuthenticate.challenge.__value.otherChallenge.authParams;
        }
      }
    else
      {
      msgHeader->proxyAuthenticate.challenge.present = TTCN_constant_Challenge_otherChallenge;
      memset(msgHeader->proxyAuthenticate.challenge.__value.otherChallenge.authScheme, 0, sizeof(RTDS_String));
      params = &msgHeader->proxyAuthenticate.challenge.__value.otherChallenge.authParams;
      }
    RTDS_CommaParam_List_init(params, TRUE);
    int i = 0;
    while (header->au_params[i])
      {
      RTDS_CommaParam_List_add_index(params, i, NULL);
      char id[256], value[256];
      int p = sscanf(header->au_params[i], "%[^=]=%[^=]", id, value);
      ((GenericParam *)params->elements)[i].idPresent = (RTDS_BOOLEAN) (p > 0);
      if (((GenericParam *)params->elements)[i].idPresent == TRUE)
        {
        strcpy(((GenericParam *)params->elements)[i].id, id);
        }
      ((GenericParam *)params->elements)[i].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
      if (((GenericParam *)params->elements)[i].paramValuePresent == TRUE)
        {
        strcpy(((GenericParam *)params->elements)[i].paramValue, value);
        }
      ++i;
      header = header->au_next;
      }
    }
  }

static void e_sip_proxy_authenticate(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->proxyAuthenticatePresent == TRUE)
    {
    strcat(buffer, "Proxy-Authenticate: ");
    CommaParam_List *params;
    if (msgHeader->proxyAuthenticate.challenge.present == TTCN_constant_Challenge_digestCln)
      {
      strcat(buffer, "Digest ");
      params = &msgHeader->proxyAuthenticate.challenge.__value.digestCln;
      }
    else
      {
      if (!strlen(msgHeader->proxyAuthenticate.challenge.__value.otherChallenge.authScheme))
        {
        strcat(buffer, msgHeader->proxyAuthenticate.challenge.__value.otherChallenge.authScheme);
        }
      params = &msgHeader->proxyAuthenticate.challenge.__value.otherChallenge.authParams;
      }
    int i;
    for (i=0; i<params->length; ++i)
      {
      if (((GenericParam *)params->elements)[i].idPresent == TRUE)
        {
        if ((i == 0 && !strlen(msgHeader->proxyAuthenticate.challenge.__value.otherChallenge.authScheme)) || i > 0)
          {
          strcat(buffer, ",");
          }
        strcat(buffer, ((GenericParam *)params->elements)[i].id);
        if (((GenericParam *)params->elements)[i].paramValuePresent == TRUE)
          {
          strcat(buffer, "=");
          strcat(buffer, ((GenericParam *)params->elements)[i].paramValue);
          }
        }
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_proxy_authorization(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_proxy_authorization_t *header = sip->sip_proxy_authorization;
  msgHeader->proxyAuthorizationPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->proxyAuthorizationPresent == TRUE)
    {
    msgHeader->proxyAuthorization.fieldName = TTCN_constant_PROXY_AUTHORIZATION_E;
    sip_authorization_e(msgHeader->proxyAuthorization.credentials, sizeof(RTDS_String), (msg_header_t *) header, 0);
    }
  }

static void e_sip_proxy_authorization(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->proxyAuthorizationPresent == TRUE)
    {
    strcat(buffer, "Proxy-Authorization: ");
    strcat(buffer, msgHeader->proxyAuthorization.credentials);
    strcat(buffer, "\r\n");
    }
  }
  
static void d_sip_proxy_require(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_proxy_require_t *header = sip->sip_proxy_require;
  msgHeader->proxyRequirePresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->proxyRequirePresent == TRUE)
    {
    msgHeader->proxyRequire.fieldName = TTCN_constant_PROXY_REQUIRE_E;
    RTDS_OptionTag_List_init(&(msgHeader->proxyRequire.optionsTags), TRUE);
    int i = 0;
    while (header != NULL)
      {
      if (header->k_items != NULL)
        {
        int j = 0;
        while (header->k_items[j])
          {
          RTDS_OptionTag_List_add_index(&(msgHeader->proxyRequire.optionsTags), i, NULL);
          strcpy(((OptionTag *)msgHeader->proxyRequire.optionsTags.elements)[i], header->k_items[j]);
          ++j;
          ++i;
          }
        }
      header = header->k_next;
      }
    }
  }

static void e_sip_proxy_require(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->proxyRequirePresent == TRUE)
    {
    strcat(buffer, "Proxy-Require: ");
    int i;
    for (i=0; i<msgHeader->proxyRequire.optionsTags.length; ++i)
      {
      if (i > 0)
        {
        strcat(buffer, ",");
        }
      strcat(buffer, ((OptionTag *)msgHeader->proxyRequire.optionsTags.elements)[i]);
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_record_route(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_record_route_t *header = sip->sip_record_route;
  msgHeader->recordRoutePresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->recordRoutePresent == TRUE)
    {
    msgHeader->recordRoute.fieldName = TTCN_constant_RECORD_ROUTE_E;
    RTDS_RouteBody_List_init(&(msgHeader->recordRoute.routeBody), TRUE);
    int i = 0;
    while (header != NULL)
      {
      RTDS_RouteBody_List_add(&(msgHeader->recordRoute.routeBody), i, NULL);
      ((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].nameAddr.displayNamePresent = (RTDS_BOOLEAN) (header->r_display != NULL);
      if (((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].nameAddr.displayNamePresent == TRUE)
        {
        strcpy(((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].nameAddr.displayName, header->r_display);
        }
      SipUrl *url = &((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].nameAddr.addrSpec;
      if (header->r_url->url_scheme != NULL)
        {
        strcpy(url->scheme, header->r_url->url_scheme);
        }
      url->userInfo.userOrTelephoneSubscriberPresent = (RTDS_BOOLEAN) (header->r_url->url_user != NULL);
      if (url->userInfo.userOrTelephoneSubscriberPresent == TRUE)
        {
        strcpy(url->userInfo.userOrTelephoneSubscriber, header->r_url->url_user);
        }
      url->userInfo.passwordPresent = (RTDS_BOOLEAN) (header->r_url->url_password != NULL);
      if (url->userInfo.passwordPresent == TRUE)
        {
        strcpy(url->userInfo.password, header->r_url->url_password);
        }
      url->userInfoPresent = (RTDS_BOOLEAN) (url->userInfo.userOrTelephoneSubscriberPresent == TRUE || url->userInfo.passwordPresent == TRUE);
      url->hostPort.hostPresent = (RTDS_BOOLEAN) (header->r_url->url_host != NULL);
      if (url->hostPort.hostPresent == TRUE)
        {
        strcpy(url->hostPort.host, header->r_url->url_host);
        }
      url->hostPort.portFieldPresent = (RTDS_BOOLEAN) (header->r_url->url_port != NULL);
      if (url->hostPort.portFieldPresent == TRUE)
        {
        url->hostPort.portField = atoi(header->r_url->url_port);
        }
      url->urlParametersPresent = (RTDS_BOOLEAN) (header->r_url->url_params != NULL);
      if (url->urlParametersPresent == TRUE)
        {
        RTDS_SemicolonParam_List_init(&(url->urlParameters), TRUE);
        int j = 0;
        char *param = strtok((char *) header->r_url->url_params, ";");
        while (param != NULL)
          {
          RTDS_SemicolonParam_List_add_index(&(url->urlParameters), j, NULL);
          char id[256], value[256];
          int p = sscanf(param, "%[^=]=%[^=]", id, value);
          ((GenericParam *)url->urlParameters.elements)[j].idPresent = (RTDS_BOOLEAN) (p > 0);
          if (((GenericParam *)url->urlParameters.elements)[j].idPresent == TRUE)
            {
            strcpy(((GenericParam *)url->urlParameters.elements)[j].id, id);
            }
          ((GenericParam *)url->urlParameters.elements)[j].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
          if (((GenericParam *)url->urlParameters.elements)[j].paramValuePresent == TRUE)
            {
            strcpy(((GenericParam *)url->urlParameters.elements)[j].paramValue, value);
            }
          ++j;
          param = strtok(NULL, ";");
          }
        }
      url->headersPresent = (RTDS_BOOLEAN) (header->r_url->url_headers != NULL);
      if (url->headersPresent == TRUE)
        {
        RTDS_AmpersandParam_List_init(&(url->headers), TRUE);
        int j = 0;
        char *param = strtok((char *) header->r_url->url_headers, "&");
        while (param != NULL)
          {
          RTDS_AmpersandParam_List_add_index(&(url->headers), j, NULL);
          char id[256], value[256];
          int p = sscanf(param, "%[^=]=%[^=]", id, value);
          ((GenericParam *)url->headers.elements)[j].idPresent = (RTDS_BOOLEAN) (p > 0);
          if (((GenericParam *)url->headers.elements)[j].idPresent == TRUE)
            {
            strcpy(((GenericParam *)url->headers.elements)[j].id, id);
            }
          ((GenericParam *)url->headers.elements)[j].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
          if (((GenericParam *)url->headers.elements)[j].paramValuePresent == TRUE)
            {
            strcpy(((GenericParam *)url->headers.elements)[j].paramValue, value);
            }
          ++j;
          param = strtok(NULL, "&");
          }
        }
      ((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].rrParamPresent = (RTDS_BOOLEAN) (header->r_params != NULL);
      if (((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].rrParamPresent == TRUE)
        {
        RTDS_SemicolonParam_List_init(&(((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].rrParam), TRUE);
        int j = 0;
        while (header->r_params[j])
          {
          RTDS_SemicolonParam_List_add_index(&(((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].rrParam), j, NULL);
          char id[256], value[256];
          int p = sscanf(header->r_params[j], "%[^=]=%[^=]", id, value);
          ((GenericParam *)((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].rrParam.elements)[j].idPresent = (RTDS_BOOLEAN) (p > 0);
          if (((GenericParam *)((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].rrParam.elements)[j].idPresent == TRUE)
            {
            strcpy(((GenericParam *)((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].rrParam.elements)[j].id, id);
            }
          ((GenericParam *)((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].rrParam.elements)[j].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
          if (((GenericParam *)((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].rrParam.elements)[j].paramValuePresent == TRUE)
            {
            strcpy(((GenericParam *)((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].rrParam.elements)[j].paramValue, value);
            }
          ++j;
          }
        }
      ++i;
      header = header->r_next;
      }
    }
  }

static void e_sip_record_route(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->recordRoutePresent == TRUE)
    {
    strcat(buffer, "Record-Route: ");
    int i, j;
    for (i=0; i<msgHeader->recordRoute.routeBody.length; ++i)
      {
      if (i > 0)
        {
        strcat(buffer, ",");
        }
      if (((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].nameAddr.displayNamePresent == TRUE)
        {
        strcat(buffer, ((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].nameAddr.displayName);
        }
      strcat(buffer, "<");
      SipUrl *url = &((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].nameAddr.addrSpec;
      if (strcmp(url->scheme, ""))
        {
        strcat(buffer, url->scheme);
        strcat(buffer, ":");
        }
      if (url->userInfo.userOrTelephoneSubscriberPresent == TRUE)
        {
        strcat(buffer, url->userInfo.userOrTelephoneSubscriber);
        if (url->userInfo.passwordPresent == TRUE)
          {
          strcat(buffer, ":");
          strcat(buffer, url->userInfo.password);
          }
        strcat(buffer, "@");
        }
      if (url->hostPort.hostPresent == TRUE)
        {
        strcat(buffer, url->hostPort.host);
        if (url->hostPort.portFieldPresent == TRUE)
          {
          strcat(buffer, ":");
          char port[8];
          sprintf(port, "%d", url->hostPort.portField);
          strcat(buffer, port);
          }
        }
      if (url->urlParametersPresent == TRUE)
        {
        for (j=0; j<url->urlParameters.length; ++j)
          {
          if (((GenericParam *)url->urlParameters.elements)[j].idPresent == TRUE)
            {
            strcat(buffer, ";");
            strcat(buffer, ((GenericParam *)url->urlParameters.elements)[j].id);
            if (((GenericParam *)url->urlParameters.elements)[j].paramValuePresent == TRUE)
              {
              strcat(buffer, "=");
              strcat(buffer, ((GenericParam *)url->urlParameters.elements)[j].paramValue);
              }
            }
          }
        }
      if (url->headersPresent == TRUE)
        {
        for(j=0; j<url->headers.length; ++j)
          {
          if (((GenericParam *)url->headers.elements)[j].idPresent == TRUE)
            {
            if (j == 0)
              {
              strcat(buffer, "?");
              }
            else
              {
              strcat(buffer, "&");
              }
            strcat(buffer, ((GenericParam *)url->headers.elements)[j].id);
            if (((GenericParam *)url->headers.elements)[j].paramValuePresent == TRUE)
              {
              strcat(buffer, "=");
              strcat(buffer, ((GenericParam *)url->headers.elements)[j].paramValue);
              }
            }
          }
        }
      strcat(buffer, ">");
      if (((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].rrParamPresent == TRUE)
        {
        for (j=0; j<((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].rrParam.length; ++j)
          {
          if (((GenericParam *)((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].rrParam.elements)[j].idPresent == TRUE)
            {
            strcat(buffer, ";");
            strcat(buffer, ((GenericParam *)((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].rrParam.elements)[j].id);
            if (((GenericParam *)((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].rrParam.elements)[j].paramValuePresent == TRUE)
              {
              strcat(buffer, "=");
              strcat(buffer, ((GenericParam *)((RouteBody *)msgHeader->recordRoute.routeBody.elements)[i].rrParam.elements)[j].paramValue);
              }
            }
          }
        }
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_reply_to(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_reply_to_t *header = sip_reply_to(sip);
  msgHeader->replyToPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->replyToPresent == TRUE)
    {
    msgHeader->replyTo.fieldName = TTCN_constant_REPLY_TO_E;
    SipUrl *url;
    if (header->rplyto_display != NULL)
      {
      msgHeader->replyTo.addressField.present = TTCN_constant_Addr_Union_nameAddr;
      msgHeader->replyTo.addressField.__value.nameAddr.displayNamePresent = TRUE;
      strcpy(msgHeader->replyTo.addressField.__value.nameAddr.displayName, header->rplyto_display);
      url = &msgHeader->replyTo.addressField.__value.nameAddr.addrSpec;
      }
    else
      {
      msgHeader->replyTo.addressField.present = TTCN_constant_Addr_Union_addrSpecUnion;
      url = &msgHeader->replyTo.addressField.__value.addrSpecUnion;
      }
    if (header->rplyto_url->url_scheme != NULL)
      {
      strcpy(url->scheme, header->rplyto_url->url_scheme);
      }
    url->userInfo.userOrTelephoneSubscriberPresent = (RTDS_BOOLEAN) (header->rplyto_url->url_user != NULL);
    if (url->userInfo.userOrTelephoneSubscriberPresent == TRUE)
      {
      strcpy(url->userInfo.userOrTelephoneSubscriber, header->rplyto_url->url_user);
      }
    url->userInfo.passwordPresent = (RTDS_BOOLEAN) (header->rplyto_url->url_password != NULL);
    if (url->userInfo.passwordPresent == TRUE)
      {
      strcpy(url->userInfo.password, header->rplyto_url->url_password);
      }
    url->userInfoPresent = (RTDS_BOOLEAN) (url->userInfo.userOrTelephoneSubscriberPresent == TRUE || url->userInfo.passwordPresent == TRUE);
    url->hostPort.hostPresent = (RTDS_BOOLEAN) (header->rplyto_url->url_host != NULL);
    if (url->hostPort.hostPresent == TRUE)
      {
      strcpy(url->hostPort.host, header->rplyto_url->url_host);
      }
    url->hostPort.portFieldPresent = (RTDS_BOOLEAN) (header->rplyto_url->url_port != NULL);
    if (url->hostPort.portFieldPresent == TRUE)
      {
      url->hostPort.portField = atoi(header->rplyto_url->url_port);
      }
    url->urlParametersPresent = (RTDS_BOOLEAN) (header->rplyto_url->url_params != NULL);
    if (url->urlParametersPresent == TRUE)
      {
      RTDS_SemicolonParam_List_init(&(url->urlParameters), TRUE);
      int i = 0;
      char *param = strtok((char *) header->rplyto_url->url_params, ";");
      while (param != NULL)
        {
        RTDS_SemicolonParam_List_add_index(&(url->urlParameters), i, NULL);
        char id[256], value[256];
        int p = sscanf(param, "%[^=]=%[^=]", id, value);
        ((GenericParam *)url->urlParameters.elements)[i].idPresent = (RTDS_BOOLEAN) (p > 0);
        if (((GenericParam *)url->urlParameters.elements)[i].idPresent == TRUE)
          {
          strcpy(((GenericParam *)url->urlParameters.elements)[i].id, id);
          }
        ((GenericParam *)url->urlParameters.elements)[i].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
        if (((GenericParam *)url->urlParameters.elements)[i].paramValuePresent == TRUE)
          {
          strcpy(((GenericParam *)url->urlParameters.elements)[i].paramValue, value);
          }
        ++i;
        param = strtok(NULL, ";");
        }
      }
    url->headersPresent = (RTDS_BOOLEAN) (header->rplyto_url->url_headers != NULL);
    if (url->headersPresent == TRUE)
      {
      RTDS_AmpersandParam_List_init(&(url->headers), TRUE);
      int i = 0;
      char *param = strtok((char *) header->rplyto_url->url_headers, "&");
      while (param != NULL)
        {
        RTDS_AmpersandParam_List_add_index(&(url->headers), i, NULL);
        char id[256], value[256];
        int p = sscanf(param, "%[^=]=%[^=]", id, value);
        ((GenericParam *)url->headers.elements)[i].idPresent = (RTDS_BOOLEAN) (p > 0);
        if (((GenericParam *)url->headers.elements)[i].idPresent == TRUE)
          {
          strcpy(((GenericParam *)url->headers.elements)[i].id, id);
          }
        ((GenericParam *)url->headers.elements)[i].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
        if (((GenericParam *)url->headers.elements)[i].paramValuePresent == TRUE)
          {
          strcpy(((GenericParam *)url->headers.elements)[i].paramValue, value);
          }
        ++i;
        param = strtok(NULL, "&");
        }
      }
    msgHeader->replyTo.replyToParamsPresent = (RTDS_BOOLEAN) (header->rplyto_params != NULL);
    if (msgHeader->replyTo.replyToParamsPresent == TRUE)
      {
      RTDS_SemicolonParam_List_init(&(msgHeader->replyTo.replyToParams), TRUE);
      int i = 0;
      while (header->rplyto_params[i])
        {
        RTDS_SemicolonParam_List_add_index(&(msgHeader->replyTo.replyToParams), i, NULL);
        char id[256], value[256];
        int p = sscanf(header->rplyto_params[i], "%[^=]=%[^=]", id, value);
        ((GenericParam *)msgHeader->replyTo.replyToParams.elements)[i].idPresent = (RTDS_BOOLEAN) (p > 0);
        if (((GenericParam *)msgHeader->replyTo.replyToParams.elements)[i].idPresent == TRUE)
          {
          strcpy(((GenericParam *)msgHeader->replyTo.replyToParams.elements)[i].id, id);
          }
        ((GenericParam *)msgHeader->replyTo.replyToParams.elements)[i].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
        if (((GenericParam *)msgHeader->replyTo.replyToParams.elements)[i].paramValuePresent == TRUE)
          {
          strcpy(((GenericParam *)msgHeader->replyTo.replyToParams.elements)[i].paramValue, value);
          }
        ++i;
        }
      }
    }
  }

static void e_sip_reply_to(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->replyToPresent == TRUE)
    {
    strcat(buffer, "Reply-To: ");
    int i;
    SipUrl *url;
    if (msgHeader->replyTo.addressField.present == TTCN_constant_Addr_Union_nameAddr)
      {
      strcat(buffer, msgHeader->replyTo.addressField.__value.nameAddr.displayName);
      strcat(buffer, "<");
      url = &msgHeader->replyTo.addressField.__value.nameAddr.addrSpec;
      }
    else
      {
      url = &msgHeader->replyTo.addressField.__value.addrSpecUnion;
      }
    if (strcmp(url->scheme, ""))
      {
      strcat(buffer, url->scheme);
      strcat(buffer, ":");
      }
    if (url->userInfo.userOrTelephoneSubscriberPresent == TRUE)
      {
      strcat(buffer, url->userInfo.userOrTelephoneSubscriber);
      if (url->userInfo.passwordPresent == TRUE)
        {
        strcat(buffer, ":");
        strcat(buffer, url->userInfo.password);
        }
      strcat(buffer, "@");
      }
    if (url->hostPort.hostPresent == TRUE)
      {
      strcat(buffer, url->hostPort.host);
      if (url->hostPort.portFieldPresent == TRUE)
        {
        strcat(buffer, ":");
        char port[8];
        sprintf(port, "%d", url->hostPort.portField);
        strcat(buffer, port);
        }
      }
    if (url->urlParametersPresent == TRUE)
      {
      for (i=0; i<url->urlParameters.length; ++i)
        {
        if (((GenericParam *)url->urlParameters.elements)[i].idPresent == TRUE)
          {
          strcat(buffer, ";");
          strcat(buffer, ((GenericParam *)url->urlParameters.elements)[i].id);
          if (((GenericParam *)url->urlParameters.elements)[i].paramValuePresent == TRUE)
            {
            strcat(buffer, "=");
            strcat(buffer, ((GenericParam *)url->urlParameters.elements)[i].paramValue);
            }
          }
        }
      }
    if (url->headersPresent == TRUE)
      {
      for(i=0; i<url->headers.length; ++i)
        {
        if (((GenericParam *)url->headers.elements)[i].idPresent == TRUE)
          {
          if (i == 0)
            {
            strcat(buffer, "?");
            }
          else
            {
            strcat(buffer, "&");
            }
          strcat(buffer, ((GenericParam *)url->headers.elements)[i].id);
          if (((GenericParam *)url->headers.elements)[i].paramValuePresent == TRUE)
            {
            strcat(buffer, "=");
            strcat(buffer, ((GenericParam *)url->headers.elements)[i].paramValue);
            }
          }
        }
      }
    if (msgHeader->replyTo.addressField.present == TTCN_constant_Addr_Union_nameAddr)
      {
      strcat(buffer, ">");
      }
    if (msgHeader->replyTo.replyToParamsPresent == TRUE)
      {
      for (i=0; i<msgHeader->replyTo.replyToParams.length; ++i)
        {
        if (((GenericParam *)msgHeader->replyTo.replyToParams.elements)[i].idPresent == TRUE)
          {
          strcat(buffer, ";");
          strcat(buffer, ((GenericParam *)msgHeader->replyTo.replyToParams.elements)[i].id);
          if (((GenericParam *)msgHeader->replyTo.replyToParams.elements)[i].paramValuePresent == TRUE)
            {
            strcat(buffer, "=");
            strcat(buffer, ((GenericParam *)msgHeader->replyTo.replyToParams.elements)[i].paramValue);
            }
          }
        }
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_require(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_require_t *header = sip->sip_require;
  msgHeader->requirePresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->requirePresent == TRUE)
    {
    msgHeader->require.fieldName = TTCN_constant_REQUIRE_E;
    RTDS_OptionTag_List_init(&(msgHeader->require.optionsTags), TRUE);
    int i = 0;
    while (header != NULL)
      {
      if (header->k_items != NULL)
        {
        int j = 0;
        while (header->k_items[j])
          {
          RTDS_OptionTag_List_add_index(&(msgHeader->require.optionsTags), i, NULL);
          strcpy(((OptionTag *)msgHeader->require.optionsTags.elements)[i], header->k_items[j]);
          ++j;
          ++i;
          }
        }
      header = header->k_next;
      }
    }
  }  

static void e_sip_require(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->requirePresent == TRUE)
    {
    strcat(buffer, "Require: ");
    int i;
    for (i=0; i<msgHeader->require.optionsTags.length; ++i)
      {
      if (i > 0)
        {
        strcat(buffer, ",");
        }
      strcat(buffer, ((OptionTag *)msgHeader->require.optionsTags.elements)[i]);
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_retry_after(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_retry_after_t *header = sip->sip_retry_after;
  msgHeader->retryAfterPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->retryAfterPresent == TRUE)
    {
    msgHeader->retryAfter.fieldName = TTCN_constant_RETRY_AFTER_E;
    sip_date_e(msgHeader->retryAfter.deltaSec, sizeof(RTDS_String), (msg_header_t *) &(header->af_delta), 0);
    msgHeader->retryAfter.commentPresent = (RTDS_BOOLEAN) (header->af_comment != NULL);
    if (msgHeader->retryAfter.commentPresent == TRUE)
      {
      strcpy(msgHeader->retryAfter.comment, header->af_comment);
      }
    msgHeader->retryAfter.retryParamsPresent = (RTDS_BOOLEAN) (header->af_params != NULL);
    if (msgHeader->retryAfter.retryParamsPresent == TRUE)
      {
      RTDS_SemicolonParam_List_init(&(msgHeader->retryAfter.retryParams), TRUE);
      int i = 0;
      while (header->af_params[i])
        {
        RTDS_SemicolonParam_List_add_index(&(msgHeader->retryAfter.retryParams), i, NULL);
        char id[256], value[256];
        int p = sscanf(header->af_params[i], "%[^=]=%[^=]", id, value);
        ((GenericParam *)msgHeader->retryAfter.retryParams.elements)[i].idPresent = (RTDS_BOOLEAN) (p > 0);
        if (((GenericParam *)msgHeader->retryAfter.retryParams.elements)[i].idPresent == TRUE)
          {
          strcpy(((GenericParam *)msgHeader->retryAfter.retryParams.elements)[i].id, id);
          }
        ((GenericParam *)msgHeader->retryAfter.retryParams.elements)[i].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
        if (((GenericParam *)msgHeader->retryAfter.retryParams.elements)[i].paramValuePresent == TRUE)
          {
          strcpy(((GenericParam *)msgHeader->retryAfter.retryParams.elements)[i].paramValue, value);
          }
        ++i;
        }
      }
    }
  }

static void e_sip_retry_after(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->retryAfterPresent == TRUE)
    {
    strcat(buffer, "Retry-After: ");
    strcat(buffer, msgHeader->retryAfter.deltaSec);
    if (msgHeader->retryAfter.commentPresent == TRUE)
      {
      strcat(buffer, "(");
      strcat(buffer, msgHeader->retryAfter.comment);
      strcat(buffer, ")");
      }
    int i;
    if (msgHeader->retryAfter.retryParamsPresent == TRUE)
      {
      for (i=0; i<msgHeader->retryAfter.retryParams.length; ++i)
        {
        if (((GenericParam *)msgHeader->retryAfter.retryParams.elements)[i].idPresent == TRUE)
          {
          strcat(buffer, ";");
          strcat(buffer, ((GenericParam *)msgHeader->retryAfter.retryParams.elements)[i].id);
          if (((GenericParam *)msgHeader->retryAfter.retryParams.elements)[i].paramValuePresent == TRUE)
            {
            strcat(buffer, "=");
            strcat(buffer, ((GenericParam *)msgHeader->retryAfter.retryParams.elements)[i].paramValue);
            }
          }
        }
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_route(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_route_t *header = sip->sip_route;
  msgHeader->routePresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->routePresent == TRUE)
    {
    msgHeader->route.fieldName = TTCN_constant_ROUTE_E;
    RTDS_RouteBody_List_init(&(msgHeader->route.routeBody), TRUE);
    int i = 0;
    while (header != NULL)
      {
      RTDS_RouteBody_List_add(&(msgHeader->route.routeBody), i, NULL);
      ((RouteBody *)msgHeader->route.routeBody.elements)[i].nameAddr.displayNamePresent = (RTDS_BOOLEAN) (header->r_display != NULL);
      if (((RouteBody *)msgHeader->route.routeBody.elements)[i].nameAddr.displayNamePresent == TRUE)
        {
        strcpy(((RouteBody *)msgHeader->route.routeBody.elements)[i].nameAddr.displayName, header->r_display);
        }
      SipUrl *url = &((RouteBody *)msgHeader->route.routeBody.elements)[i].nameAddr.addrSpec;
      if (header->r_url->url_scheme != NULL)
        {
        strcpy(url->scheme, header->r_url->url_scheme);
        }
      url->userInfo.userOrTelephoneSubscriberPresent = (RTDS_BOOLEAN) (header->r_url->url_user != NULL);
      if (url->userInfo.userOrTelephoneSubscriberPresent == TRUE)
        {
        strcpy(url->userInfo.userOrTelephoneSubscriber, header->r_url->url_user);
        }
      url->userInfo.passwordPresent = (RTDS_BOOLEAN) (header->r_url->url_password != NULL);
      if (url->userInfo.passwordPresent == TRUE)
        {
        strcpy(url->userInfo.password, header->r_url->url_password);
        }
      url->userInfoPresent = (RTDS_BOOLEAN) (url->userInfo.userOrTelephoneSubscriberPresent == TRUE || url->userInfo.passwordPresent == TRUE);
      url->hostPort.hostPresent = (RTDS_BOOLEAN) (header->r_url->url_host != NULL);
      if (url->hostPort.hostPresent == TRUE)
        {
        strcpy(url->hostPort.host, header->r_url->url_host);
        }
      url->hostPort.portFieldPresent = (RTDS_BOOLEAN) (header->r_url->url_port != NULL);
      if (url->hostPort.portFieldPresent == TRUE)
        {
        url->hostPort.portField = atoi(header->r_url->url_port);
        }
      url->urlParametersPresent = (RTDS_BOOLEAN) (header->r_url->url_params != NULL);
      if (url->urlParametersPresent == TRUE)
        {
        RTDS_SemicolonParam_List_init(&(url->urlParameters), TRUE);
        int j = 0;
        char *param = strtok((char *) header->r_url->url_params, ";");
        while (param != NULL)
          {
          RTDS_SemicolonParam_List_add_index(&(url->urlParameters), j, NULL);
          char id[256], value[256];
          int p = sscanf(param, "%[^=]=%[^=]", id, value);
          ((GenericParam *)url->urlParameters.elements)[j].idPresent = (RTDS_BOOLEAN) (p > 0);
          if (((GenericParam *)url->urlParameters.elements)[j].idPresent == TRUE)
            {
            strcpy(((GenericParam *)url->urlParameters.elements)[j].id, id);
            }
          ((GenericParam *)url->urlParameters.elements)[j].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
          if (((GenericParam *)url->urlParameters.elements)[j].paramValuePresent == TRUE)
            {
            strcpy(((GenericParam *)url->urlParameters.elements)[j].paramValue, value);
            }
          ++j;
          param = strtok(NULL, ";");
          }
        }
      url->headersPresent = (RTDS_BOOLEAN) (header->r_url->url_headers != NULL);
      if (url->headersPresent == TRUE)
        {
        RTDS_AmpersandParam_List_init(&(url->headers), TRUE);
        int j = 0;
        char *param = strtok((char *) header->r_url->url_headers, "&");
        while (param != NULL)
          {
          RTDS_AmpersandParam_List_add_index(&(url->headers), j, NULL);
          char id[256], value[256];
          int p = sscanf(param, "%[^=]=%[^=]", id, value);
          ((GenericParam *)url->headers.elements)[j].idPresent = (RTDS_BOOLEAN) (p > 0);
          if (((GenericParam *)url->headers.elements)[j].idPresent == TRUE)
            {
            strcpy(((GenericParam *)url->headers.elements)[j].id, id);
            }
          ((GenericParam *)url->headers.elements)[j].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
          if (((GenericParam *)url->headers.elements)[j].paramValuePresent == TRUE)
            {
            strcpy(((GenericParam *)url->headers.elements)[j].paramValue, value);
            }
          ++j;
          param = strtok(NULL, "&");
          }
        }
      ((RouteBody *)msgHeader->route.routeBody.elements)[i].rrParamPresent = (RTDS_BOOLEAN) (header->r_params != NULL);
      if (((RouteBody *)msgHeader->route.routeBody.elements)[i].rrParamPresent == TRUE)
        {
        RTDS_SemicolonParam_List_init(&(((RouteBody *)msgHeader->route.routeBody.elements)[i].rrParam), TRUE);
        int j = 0;
        while (header->r_params[j])
          {
          RTDS_SemicolonParam_List_add_index(&(((RouteBody *)msgHeader->route.routeBody.elements)[i].rrParam), j, NULL);
          char id[256], value[256];
          int p = sscanf(header->r_params[j], "%[^=]=%[^=]", id, value);
          ((GenericParam *)((RouteBody *)msgHeader->route.routeBody.elements)[i].rrParam.elements)[j].idPresent = (RTDS_BOOLEAN) (p > 0);
          if (((GenericParam *)((RouteBody *)msgHeader->route.routeBody.elements)[i].rrParam.elements)[j].idPresent == TRUE)
            {
            strcpy(((GenericParam *)((RouteBody *)msgHeader->route.routeBody.elements)[i].rrParam.elements)[j].id, id);
            }
          ((GenericParam *)((RouteBody *)msgHeader->route.routeBody.elements)[i].rrParam.elements)[j].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
          if (((GenericParam *)((RouteBody *)msgHeader->route.routeBody.elements)[i].rrParam.elements)[j].paramValuePresent == TRUE)
            {
            strcpy(((GenericParam *)((RouteBody *)msgHeader->route.routeBody.elements)[i].rrParam.elements)[j].paramValue, value);
            }
          ++j;
          }
        }
      ++i;
      header = header->r_next;
      }
    }
  }

static void e_sip_route(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->routePresent == TRUE)
    {
    strcat(buffer, "Route: ");
    int i, j;
    for (i=0; i<msgHeader->route.routeBody.length; ++i)
      {
      if (i > 0)
        {
        strcat(buffer, ",");
        }
      if (((RouteBody *)msgHeader->route.routeBody.elements)[i].nameAddr.displayNamePresent == TRUE)
        {
        strcat(buffer, ((RouteBody *)msgHeader->route.routeBody.elements)[i].nameAddr.displayName);
        }
      strcat(buffer, "<");
      SipUrl *url = &((RouteBody *)msgHeader->route.routeBody.elements)[i].nameAddr.addrSpec;
      if (strcmp(url->scheme, ""))
        {
        strcat(buffer, url->scheme);
        strcat(buffer, ":");
        }
      if (url->userInfo.userOrTelephoneSubscriberPresent == TRUE)
        {
        strcat(buffer, url->userInfo.userOrTelephoneSubscriber);
        if (url->userInfo.passwordPresent == TRUE)
          {
          strcat(buffer, ":");
          strcat(buffer, url->userInfo.password);
          }
        strcat(buffer, "@");
        }
      if (url->hostPort.hostPresent == TRUE)
        {
        strcat(buffer, url->hostPort.host);
        if (url->hostPort.portFieldPresent == TRUE)
          {
          strcat(buffer, ":");
          char port[8];
          sprintf(port, "%d", url->hostPort.portField);
          strcat(buffer, port);
          }
        }
      if (url->urlParametersPresent == TRUE)
        {
        for (j=0; j<url->urlParameters.length; ++j)
          {
          if (((GenericParam *)url->urlParameters.elements)[j].idPresent == TRUE)
            {
            strcat(buffer, ";");
            strcat(buffer, ((GenericParam *)url->urlParameters.elements)[j].id);
            if (((GenericParam *)url->urlParameters.elements)[j].paramValuePresent == TRUE)
              {
              strcat(buffer, "=");
              strcat(buffer, ((GenericParam *)url->urlParameters.elements)[j].paramValue);
              }
            }
          }
        }
      if (url->headersPresent == TRUE)
        {
        for(j=0; j<url->headers.length; ++j)
          {
          if (((GenericParam *)url->headers.elements)[j].idPresent == TRUE)
            {
            if (j == 0)
              {
              strcat(buffer, "?");
              }
            else
              {
              strcat(buffer, "&");
              }
            strcat(buffer, ((GenericParam *)url->headers.elements)[j].id);
            if (((GenericParam *)url->headers.elements)[j].paramValuePresent == TRUE)
              {
              strcat(buffer, "=");
              strcat(buffer, ((GenericParam *)url->headers.elements)[j].paramValue);
              }
            }
          }
        }
      strcat(buffer, ">");
      if (((RouteBody *)msgHeader->route.routeBody.elements)[i].rrParamPresent == TRUE)
        {
        for (j=0; j<((RouteBody *)msgHeader->route.routeBody.elements)[i].rrParam.length; ++j)
          {
          if (((GenericParam *)((RouteBody *)msgHeader->route.routeBody.elements)[i].rrParam.elements)[j].idPresent == TRUE)
            {
            strcat(buffer, ";");
            strcat(buffer, ((GenericParam *)((RouteBody *)msgHeader->route.routeBody.elements)[i].rrParam.elements)[j].id);
            if (((GenericParam *)((RouteBody *)msgHeader->route.routeBody.elements)[i].rrParam.elements)[j].paramValuePresent == TRUE)
              {
              strcat(buffer, "=");
              strcat(buffer, ((GenericParam *)((RouteBody *)msgHeader->route.routeBody.elements)[i].rrParam.elements)[j].paramValue);
              }
            }
          }
        }
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_server(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_server_t *header = sip->sip_server;
  msgHeader->serverPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->serverPresent == TRUE)
    {
    msgHeader->server.fieldName = TTCN_constant_SERVER_E;
    RTDS_ServerVal_List_init(&(msgHeader->server.serverBody), TRUE);
    int i = 0;
    while (header != NULL)
      {
      if (header->g_string != NULL)
        {
        RTDS_ServerVal_List_add_index(&(msgHeader->server.serverBody), i, NULL);
        strcpy(((ServerVal *)msgHeader->server.serverBody.elements)[i], header->g_string);
        ++i;
        }
      header = header->g_next;
      }
    }
  }

static void e_sip_server(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->serverPresent == TRUE)
    {
    strcat(buffer, "Server: ");
    int i;
    for (i=0; i<msgHeader->server.serverBody.length; ++i)
      {
      if (i > 0)
        {
        strcat(buffer, " ");
        }
      strcat(buffer, ((ServerVal *)msgHeader->server.serverBody.elements)[i]);
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_subject(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_subject_t *header = sip->sip_subject;
  msgHeader->subjectPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->subjectPresent == TRUE)
    {
    msgHeader->subject.fieldName = TTCN_constant_SUBJECT_E;
    if (header->g_string != NULL)
      {
      strcpy(msgHeader->subject.summary, header->g_string);
      }
    else
      {
      memset(msgHeader->subject.summary, 0, sizeof(RTDS_String));
      }
    }
  }

static void e_sip_subject(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->subjectPresent == TRUE)
    {
    strcat(buffer, "Subject: ");
    if (!strlen(msgHeader->subject.summary))
      {
      strcat(buffer, msgHeader->subject.summary);
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_supported(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_supported_t *header = sip->sip_supported;
  msgHeader->supportedPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->supportedPresent == TRUE)
    {
    msgHeader->supported.fieldName = TTCN_constant_SUPPORTED_E;
    RTDS_OptionTag_List_init(&(msgHeader->supported.optionsTags), TRUE);
    int i = 0;
    while (header != NULL)
      {
      if (header->k_items != NULL)
        {
        int j = 0;
        while (header->k_items[j])
          {
          RTDS_OptionTag_List_add_index(&(msgHeader->supported.optionsTags), i, NULL);
          strcpy(((OptionTag *)msgHeader->supported.optionsTags.elements)[i], header->k_items[j]);
          ++j;
          ++i;
          }
        }
      header = header->k_next;
      }
    }
  }

static void e_sip_supported(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->supportedPresent == TRUE)
    {
    strcat(buffer, "Supported: ");
    int i;
    for (i=0; i<msgHeader->supported.optionsTags.length; ++i)
      {
      if (i > 0)
        {
        strcat(buffer, ",");
        }
      strcat(buffer, ((OptionTag *)msgHeader->supported.optionsTags.elements)[i]);
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_timestamp(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_timestamp_t *header = sip->sip_timestamp;
  msgHeader->timestampPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->timestampPresent == TRUE)
    {
    msgHeader->timestamp.fieldName = TTCN_constant_TIMESTAMP_E;
    msgHeader->timestamp.timeValuePresent = (RTDS_BOOLEAN) (header->ts_stamp != NULL);
    if (msgHeader->timestamp.timeValuePresent == TRUE)
      {
      if (sscanf(header->ts_stamp, "%d.%d", &msgHeader->timestamp.timeValue.majorDigit, &msgHeader->timestamp.timeValue.minorDigit) == 2)
        {
        msgHeader->timestamp.timeValue.minorDigitPresent = TRUE;
        }
      }
    msgHeader->timestamp.delayPresent = (RTDS_BOOLEAN) (header->ts_delay != NULL);
    if (msgHeader->timestamp.delayPresent == TRUE)
      {
      if (sscanf(header->ts_delay, "%d.%d", &msgHeader->timestamp.delay.majorDigit, &msgHeader->timestamp.delay.minorDigit) == 2)
        {
        msgHeader->timestamp.delay.minorDigitPresent = TRUE;
        }
      }
    }
  }

static void e_sip_timestamp(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->timestampPresent == TRUE)
    {
    strcat(buffer, "Timestamp: ");
    char digit[32];
    sprintf(digit, "%d", msgHeader->timestamp.timeValue.majorDigit);
    strcat(buffer, digit);
    if (msgHeader->timestamp.timeValue.minorDigitPresent == TRUE)
      {
      sprintf(digit, ".%d", msgHeader->timestamp.timeValue.minorDigit);
      strcat(buffer, digit);
      }
    if (msgHeader->timestamp.delayPresent == TRUE)
      {
      strcat(buffer, " ");
      sprintf(digit, "%d", msgHeader->timestamp.delay.majorDigit);
      strcat(buffer, digit);
      if (msgHeader->timestamp.delay.minorDigitPresent == TRUE)
        {
        sprintf(digit, ".%d", msgHeader->timestamp.delay.minorDigit);
        strcat(buffer, digit);
        }
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_to(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_to_t *header = sip->sip_to;
  msgHeader->toFieldPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->toFieldPresent == TRUE)
    {
    msgHeader->toField.fieldName = TTCN_constant_TO_E;
    SipUrl *url;
    if (header->a_display != NULL)
      {
      msgHeader->toField.addressField.present = TTCN_constant_Addr_Union_nameAddr;
      msgHeader->toField.addressField.__value.nameAddr.displayNamePresent = TRUE;
      strcpy(msgHeader->toField.addressField.__value.nameAddr.displayName, header->a_display);
      url = &msgHeader->toField.addressField.__value.nameAddr.addrSpec;
      }
    else
      {
      msgHeader->toField.addressField.present = TTCN_constant_Addr_Union_addrSpecUnion;
      url = &msgHeader->toField.addressField.__value.addrSpecUnion;
      }
    if (header->a_url->url_scheme != NULL)
      {
      strcpy(url->scheme, header->a_url->url_scheme);
      }
    url->userInfo.userOrTelephoneSubscriberPresent = (RTDS_BOOLEAN) (header->a_url->url_user != NULL);
    if (url->userInfo.userOrTelephoneSubscriberPresent == TRUE)
      {
      strcpy(url->userInfo.userOrTelephoneSubscriber, header->a_url->url_user);
      }
    url->userInfo.passwordPresent = (RTDS_BOOLEAN) (header->a_url->url_password != NULL);
    if (url->userInfo.passwordPresent == TRUE)
      {
      strcpy(url->userInfo.password, header->a_url->url_password);
      }
    url->userInfoPresent = (RTDS_BOOLEAN) (url->userInfo.userOrTelephoneSubscriberPresent == TRUE || url->userInfo.passwordPresent == TRUE);
    url->hostPort.hostPresent = (RTDS_BOOLEAN) (header->a_url->url_host != NULL);
    if (url->hostPort.hostPresent == TRUE)
      {
      strcpy(url->hostPort.host, header->a_url->url_host);
      }
    url->hostPort.portFieldPresent = (RTDS_BOOLEAN) (header->a_url->url_port != NULL);
    if (url->hostPort.portFieldPresent == TRUE)
      {
      url->hostPort.portField = atoi(header->a_url->url_port);
      }
    url->urlParametersPresent = (RTDS_BOOLEAN) (header->a_url->url_params != NULL);
    if (url->urlParametersPresent == TRUE)
      {
      RTDS_SemicolonParam_List_init(&(url->urlParameters), TRUE);
      int i = 0;
      char *param = strtok((char *) header->a_url->url_params, ";");
      while (param != NULL)
        {
        RTDS_SemicolonParam_List_add_index(&(url->urlParameters), i, NULL);
        char id[256], value[256];
        int p = sscanf(param, "%[^=]=%[^=]", id, value);
        ((GenericParam *)url->urlParameters.elements)[i].idPresent = (RTDS_BOOLEAN) (p > 0);
        if (((GenericParam *)url->urlParameters.elements)[i].idPresent == TRUE)
          {
          strcpy(((GenericParam *)url->urlParameters.elements)[i].id, id);
          }
        ((GenericParam *)url->urlParameters.elements)[i].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
        if (((GenericParam *)url->urlParameters.elements)[i].paramValuePresent == TRUE)
          {
          strcpy(((GenericParam *)url->urlParameters.elements)[i].paramValue, value);
          }
        ++i;
        param = strtok(NULL, ";");
        }
      }
    url->headersPresent = (RTDS_BOOLEAN) (header->a_url->url_headers != NULL);
    if (url->headersPresent == TRUE)
      {
      RTDS_AmpersandParam_List_init(&(url->headers), TRUE);
      int i = 0;
      char *param = strtok((char *) header->a_url->url_headers, "&");
      while (param != NULL)
        {
        RTDS_AmpersandParam_List_add_index(&(url->headers), i, NULL);
        char id[256], value[256];
        int p = sscanf(param, "%[^=]=%[^=]", id, value);
        ((GenericParam *)url->headers.elements)[i].idPresent = (RTDS_BOOLEAN) (p > 0);
        if (((GenericParam *)url->headers.elements)[i].idPresent == TRUE)
          {
          strcpy(((GenericParam *)url->headers.elements)[i].id, id);
          }
        ((GenericParam *)url->headers.elements)[i].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
        if (((GenericParam *)url->headers.elements)[i].paramValuePresent == TRUE)
          {
          strcpy(((GenericParam *)url->headers.elements)[i].paramValue, value);
          }
        ++i;
        param = strtok(NULL, "&");
        }
      }
    msgHeader->toField.toParamsPresent = (RTDS_BOOLEAN) (header->a_params != NULL);
    if (msgHeader->toField.toParamsPresent == TRUE)
      {
      RTDS_SemicolonParam_List_init(&(msgHeader->toField.toParams), TRUE);
      int i = 0;
      while (header->a_params[i])
        {
        RTDS_SemicolonParam_List_add_index(&(msgHeader->toField.toParams), i, NULL);
        char id[256], value[256];
        int p = sscanf(header->a_params[i], "%[^=]=%[^=]", id, value);
        ((GenericParam *)msgHeader->toField.toParams.elements)[i].idPresent = (RTDS_BOOLEAN) (p > 0);
        if (((GenericParam *)msgHeader->toField.toParams.elements)[i].idPresent == TRUE)
          {
          strcpy(((GenericParam *)msgHeader->toField.toParams.elements)[i].id, id);
          }
        ((GenericParam *)msgHeader->toField.toParams.elements)[i].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
        if (((GenericParam *)msgHeader->toField.toParams.elements)[i].paramValuePresent == TRUE)
          {
          strcpy(((GenericParam *)msgHeader->toField.toParams.elements)[i].paramValue, value);
          }
        ++i;
        }
      }
    }
  }

static void e_sip_to(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->toFieldPresent == TRUE)
    {
    strcat(buffer, "To: ");
    int i;
    SipUrl *url;
    if (msgHeader->toField.addressField.present == TTCN_constant_Addr_Union_nameAddr)
      {
      strcat(buffer, msgHeader->toField.addressField.__value.nameAddr.displayName);
      strcat(buffer, "<");
      url = &msgHeader->toField.addressField.__value.nameAddr.addrSpec;
      }
    else
      {
      url = &msgHeader->toField.addressField.__value.addrSpecUnion;
      }
    if (strcmp(url->scheme, ""))
      {
      strcat(buffer, url->scheme);
      strcat(buffer, ":");
      }
    if (url->userInfo.userOrTelephoneSubscriberPresent == TRUE)
      {
      strcat(buffer, url->userInfo.userOrTelephoneSubscriber);
      if (url->userInfo.passwordPresent == TRUE)
        {
        strcat(buffer, ":");
        strcat(buffer, url->userInfo.password);
        }
      strcat(buffer, "@");
      }
    if (url->hostPort.hostPresent == TRUE)
      {
      strcat(buffer, url->hostPort.host);
      if (url->hostPort.portFieldPresent == TRUE)
        {
        strcat(buffer, ":");
        char port[8];
        sprintf(port, "%d", url->hostPort.portField);
        strcat(buffer, port);
        }
      }
    if (url->urlParametersPresent == TRUE)
      {
      for (i=0; i<url->urlParameters.length; ++i)
        {
        if (((GenericParam *)url->urlParameters.elements)[i].idPresent == TRUE)
          {
          strcat(buffer, ";");
          strcat(buffer, ((GenericParam *)url->urlParameters.elements)[i].id);
          if (((GenericParam *)url->urlParameters.elements)[i].paramValuePresent == TRUE)
            {
            strcat(buffer, "=");
            strcat(buffer, ((GenericParam *)url->urlParameters.elements)[i].paramValue);
            }
          }
        }
      }
    if (url->headersPresent == TRUE)
      {
      for(i=0; i<url->headers.length; ++i)
        {
        if (((GenericParam *)url->headers.elements)[i].idPresent == TRUE)
          {
          if (i == 0)
            {
            strcat(buffer, "?");
            }
          else
            {
            strcat(buffer, "&");
            }
          strcat(buffer, ((GenericParam *)url->headers.elements)[i].id);
          if (((GenericParam *)url->headers.elements)[i].paramValuePresent == TRUE)
            {
            strcat(buffer, "=");
            strcat(buffer, ((GenericParam *)url->headers.elements)[i].paramValue);
            }
          }
        }
      }
    if (msgHeader->toField.addressField.present == TTCN_constant_Addr_Union_nameAddr)
      {
      strcat(buffer, ">");
      }
    if (msgHeader->toField.toParamsPresent == TRUE)
      {
      for (i=0; i<msgHeader->toField.toParams.length; ++i)
        {
        if (((GenericParam *)msgHeader->toField.toParams.elements)[i].idPresent == TRUE)
          {
          strcat(buffer, ";");
          strcat(buffer, ((GenericParam *)msgHeader->toField.toParams.elements)[i].id);
          if (((GenericParam *)msgHeader->toField.toParams.elements)[i].paramValuePresent == TRUE)
            {
            strcat(buffer, "=");
            strcat(buffer, ((GenericParam *)msgHeader->toField.toParams.elements)[i].paramValue);
            }
          }
        }
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_unsupported(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_unsupported_t *header = sip->sip_unsupported;
  msgHeader->unsupportedPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->unsupportedPresent == TRUE)
    {
    msgHeader->unsupported.fieldName = TTCN_constant_UNSUPPORTED_E;
    RTDS_OptionTag_List_init(&(msgHeader->unsupported.optionsTags), TRUE);
    int i = 0;
    while (header != NULL)
      {
      if (header->k_items != NULL)
        {
        int j = 0;
        while (header->k_items[j])
          {
          RTDS_OptionTag_List_add_index(&(msgHeader->unsupported.optionsTags), i, NULL);
          strcpy(((OptionTag *)msgHeader->unsupported.optionsTags.elements)[i], header->k_items[j]);
          ++j;
          ++i;
          }
        }
      header = header->k_next;
      }
    }
  }

static void e_sip_unsupported(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->unsupportedPresent == TRUE)
    {
    strcat(buffer, "Unsupported: ");
    int i;
    for (i=0; i<msgHeader->unsupported.optionsTags.length; ++i)
      {
      if (i > 0)
        {
        strcat(buffer, ",");
        }
      strcat(buffer, ((OptionTag *)msgHeader->unsupported.optionsTags.elements)[i]);
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_user_agent(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_user_agent_t *header = sip->sip_user_agent;
  msgHeader->userAgentPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->userAgentPresent == TRUE)
    {
    msgHeader->userAgent.fieldName = TTCN_constant_USER_AGENT_E;
    RTDS_ServerVal_List_init(&(msgHeader->userAgent.userAgentBody), TRUE);
    int i = 0;
    while (header != NULL)
      {
      if (header->g_string != NULL)
        {
        RTDS_ServerVal_List_add_index(&(msgHeader->userAgent.userAgentBody), i, NULL);
        strcpy(((ServerVal *)msgHeader->userAgent.userAgentBody.elements)[i], header->g_string);
        ++i;
        }
      header = header->g_next;
      }
    }
  }

static void e_sip_user_agent(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->userAgentPresent == TRUE)
    {
    strcat(buffer, "User-Agent: ");
    int i;
    for (i=0; i<msgHeader->userAgent.userAgentBody.length; ++i)
      {
      if (i > 0)
        {
        strcat(buffer, " ");
        }
      strcat(buffer, ((ServerVal *)msgHeader->userAgent.userAgentBody.elements)[i]);
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_via(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_via_t *header = sip->sip_via;
  msgHeader->viaPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->viaPresent == TRUE)
    {
    msgHeader->via.fieldName = TTCN_constant_VIA_E;
    RTDS_ViaBody_List_init(&(msgHeader->via.viaBody), TRUE);
    int i = 0;
    while (header != NULL)
      {
      RTDS_ViaBody_List_add(&(msgHeader->via.viaBody), i, NULL);
      if (header->v_protocol != NULL)
        {
        sscanf(header->v_protocol, "%[^/]/%[^/]/%[^/]", 
          ((ViaBody *)msgHeader->via.viaBody.elements)[i].sentProtocol.protocolName,
          ((ViaBody *)msgHeader->via.viaBody.elements)[i].sentProtocol.protocolVersion,
          ((ViaBody *)msgHeader->via.viaBody.elements)[i].sentProtocol.transport);
        }
      ((ViaBody *)msgHeader->via.viaBody.elements)[i].sentBy.hostPresent = (RTDS_BOOLEAN) (header->v_host != NULL);
      if (((ViaBody *)msgHeader->via.viaBody.elements)[i].sentBy.hostPresent == TRUE)
        {
        strcpy(((ViaBody *)msgHeader->via.viaBody.elements)[i].sentBy.host, header->v_host);
        }
      ((ViaBody *)msgHeader->via.viaBody.elements)[i].sentBy.portFieldPresent = (RTDS_BOOLEAN) (header->v_port != NULL);
      if (((ViaBody *)msgHeader->via.viaBody.elements)[i].sentBy.portFieldPresent == TRUE)
        {
        ((ViaBody *)msgHeader->via.viaBody.elements)[i].sentBy.portField = atoi(header->v_port);
        }
      ((ViaBody *)msgHeader->via.viaBody.elements)[i].viaParamsPresent = (RTDS_BOOLEAN) (header->v_params != NULL);
      if (((ViaBody *)msgHeader->via.viaBody.elements)[i].viaParamsPresent == TRUE)
        {
        RTDS_SemicolonParam_List_init(&(((ViaBody *)msgHeader->via.viaBody.elements)[i].viaParams), TRUE);
        int j = 0;
        while (header->v_params[j])
          {
          RTDS_SemicolonParam_List_add_index(&(((ViaBody *)msgHeader->via.viaBody.elements)[i].viaParams), j, NULL);
          char id[256], value[256];
          int p = sscanf(header->v_params[j], "%[^=]=%[^=]", id, value);
          ((GenericParam *)((ViaBody *)msgHeader->via.viaBody.elements)[i].viaParams.elements)[j].idPresent = (RTDS_BOOLEAN) (p > 0);
          if (((GenericParam *)((ViaBody *)msgHeader->via.viaBody.elements)[i].viaParams.elements)[j].idPresent == TRUE)
            {
            strcpy(((GenericParam *)((ViaBody *)msgHeader->via.viaBody.elements)[i].viaParams.elements)[j].id, id);
            }
          ((GenericParam *)((ViaBody *)msgHeader->via.viaBody.elements)[i].viaParams.elements)[j].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
          if (((GenericParam *)((ViaBody *)msgHeader->via.viaBody.elements)[i].viaParams.elements)[j].paramValuePresent == TRUE)
            {
            strcpy(((GenericParam *)((ViaBody *)msgHeader->via.viaBody.elements)[i].viaParams.elements)[j].paramValue, value);
            }
          ++j;
          }
        }
      ++i;
      header = header->v_next;
      }
    }
  }

static void e_sip_via(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->viaPresent == TRUE)
    {
    strcat(buffer, "Via: ");
    int i, j;
    for (i=0; i<msgHeader->via.viaBody.length; ++i)
      {
      if (i > 0)
        {
        strcat(buffer, ",");
        }
      strcat(buffer, ((ViaBody *)msgHeader->via.viaBody.elements)[i].sentProtocol.protocolName);
      strcat(buffer, "/");
      strcat(buffer, ((ViaBody *)msgHeader->via.viaBody.elements)[i].sentProtocol.protocolVersion);
      strcat(buffer, "/");
      strcat(buffer, ((ViaBody *)msgHeader->via.viaBody.elements)[i].sentProtocol.transport);
      strcat(buffer, " ");
      if (((ViaBody *)msgHeader->via.viaBody.elements)[i].sentBy.hostPresent == TRUE)
        {
        strcat(buffer, ((ViaBody *)msgHeader->via.viaBody.elements)[i].sentBy.host);
        if (((ViaBody *)msgHeader->via.viaBody.elements)[i].sentBy.portFieldPresent == TRUE)
          {
          char port[32];
          sprintf(port, ":%d", ((ViaBody *)msgHeader->via.viaBody.elements)[i].sentBy.portField);
          strcat(buffer, port);
          }
        }
      if (((ViaBody *)msgHeader->via.viaBody.elements)[i].viaParamsPresent == TRUE)
        {
        for (j=0; j<((ViaBody *)msgHeader->via.viaBody.elements)[i].viaParams.length; ++j)
          {
          if (((GenericParam *)((ViaBody *)msgHeader->via.viaBody.elements)[i].viaParams.elements)[j].idPresent == TRUE)
            {
            strcat(buffer, ";");
            strcat(buffer, ((GenericParam *)((ViaBody *)msgHeader->via.viaBody.elements)[i].viaParams.elements)[j].id);
            if (((GenericParam *)((ViaBody *)msgHeader->via.viaBody.elements)[i].viaParams.elements)[j].paramValuePresent == TRUE)
              {
              strcat(buffer, "=");
              strcat(buffer, ((GenericParam *)((ViaBody *)msgHeader->via.viaBody.elements)[i].viaParams.elements)[j].paramValue);
              }
            }
          }
        }
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_warning(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_warning_t *header = sip->sip_warning;
  msgHeader->warningPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->warningPresent == TRUE)
    {
    msgHeader->warning.fieldName = TTCN_constant_WARNING_E;
    RTDS_WarningValue_List_init(&(msgHeader->warning.warningValue), TRUE);
    int i = 0;
    while (header != NULL)
      {
      RTDS_WarningValue_List_add_index(&(msgHeader->warning.warningValue), i, NULL);
      ((WarningValue *)msgHeader->warning.warningValue.elements)[i].warnCode = (int) header->w_code;
      if (header->w_port != NULL)
        {
        ((WarningValue *)msgHeader->warning.warningValue.elements)[i].warnAgent.present = TTCN_constant_WarnAgent_hostPort;
        ((WarningValue *)msgHeader->warning.warningValue.elements)[i].warnAgent.__value.hostPort.portFieldPresent = TRUE;
        ((WarningValue *)msgHeader->warning.warningValue.elements)[i].warnAgent.__value.hostPort.portField = atoi(header->w_port);
        ((WarningValue *)msgHeader->warning.warningValue.elements)[i].warnAgent.__value.hostPort.hostPresent = (RTDS_BOOLEAN) (header->w_host != NULL);
        if (((WarningValue *)msgHeader->warning.warningValue.elements)[i].warnAgent.__value.hostPort.hostPresent == TRUE)
          {
          strcpy(((WarningValue *)msgHeader->warning.warningValue.elements)[i].warnAgent.__value.hostPort.host, header->w_host);
          }
        }
      else
        {
        ((WarningValue *)msgHeader->warning.warningValue.elements)[i].warnAgent.present = TTCN_constant_WarnAgent_pseudonym;
        if (header->w_host != NULL)
          {
          strcpy(((WarningValue *)msgHeader->warning.warningValue.elements)[i].warnAgent.__value.pseudonym, header->w_host);
          }
        }
      if (header->w_text != NULL)
        {
        strcpy(((WarningValue *)msgHeader->warning.warningValue.elements)[i].warnText, header->w_text);
        }
      ++i;
      header = header->w_next;
      }
    }
  }

static void e_sip_warning(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->warningPresent == TRUE)
    {
    strcat(buffer, "Warning: ");
    int i;
    for (i=0; i<msgHeader->warning.warningValue.length; ++i)
      {
      if (i > 0)
        {
        strcat(buffer, ",");
        }
      char digit[32];
      sprintf(digit, "%d", ((WarningValue *)msgHeader->warning.warningValue.elements)[i].warnCode);
      strcat(buffer, digit);
      strcat(buffer, " ");
      if (((WarningValue *)msgHeader->warning.warningValue.elements)[i].warnAgent.present == TTCN_constant_WarnAgent_hostPort)
        {
        strcat(buffer, ((WarningValue *)msgHeader->warning.warningValue.elements)[i].warnAgent.__value.hostPort.host);
        if (((WarningValue *)msgHeader->warning.warningValue.elements)[i].warnAgent.__value.hostPort.portFieldPresent == TRUE)
          {
          sprintf(digit, "%d", ((WarningValue *)msgHeader->warning.warningValue.elements)[i].warnAgent.__value.hostPort.portField);
          strcat(buffer, digit);
          }
        }
      else
        {
        strcat(buffer, ((WarningValue *)msgHeader->warning.warningValue.elements)[i].warnAgent.__value.pseudonym);
        }
      strcat(buffer, " \"");
      strcat(buffer, ((WarningValue *)msgHeader->warning.warningValue.elements)[i].warnText);
      strcat(buffer, "\"");
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_www_authenticate(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_www_authenticate_t *header = sip->sip_www_authenticate;
  msgHeader->wwwAuthenticatePresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->wwwAuthenticatePresent == TRUE)
    {
    msgHeader->wwwAuthenticate.fieldName = TTCN_constant_WWW_AUTHENTICATE_E;
    if (header->au_scheme != NULL)
      {
      CommaParam_List *params;
      if (strncmp(header->au_scheme, "Digest", 6) == 0)
        {
        msgHeader->wwwAuthenticate.challenge.present = TTCN_constant_Challenge_digestCln;
        params = &msgHeader->wwwAuthenticate.challenge.__value.digestCln;
        }
      else
        {
        msgHeader->wwwAuthenticate.challenge.present = TTCN_constant_Challenge_otherChallenge;
        strcpy(msgHeader->wwwAuthenticate.challenge.__value.otherChallenge.authScheme, header->au_scheme);
        params = &msgHeader->wwwAuthenticate.challenge.__value.otherChallenge.authParams;
        }
      RTDS_CommaParam_List_init(params, TRUE);
      int i = 0;
      while (header->au_params[i])
        {
        RTDS_CommaParam_List_add_index(params, i, NULL);
        char id[256], value[256];
        int p = sscanf(header->au_params[i], "%[^=]=%[^=]", id, value);
        ((GenericParam *)params->elements)[i].idPresent = (RTDS_BOOLEAN) (p > 0);
        if (((GenericParam *)params->elements)[i].idPresent == TRUE)
          {
          strcpy(((GenericParam *)params->elements)[i].id, id);
          }
        ((GenericParam *)params->elements)[i].paramValuePresent = (RTDS_BOOLEAN) (p == 2);
        if (((GenericParam *)params->elements)[i].paramValuePresent == TRUE)
          {
          strcpy(((GenericParam *)params->elements)[i].paramValue, value);
          }
        ++i;
        header = header->au_next;
        }
      }
    }
  }

static void e_sip_www_authenticate(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->wwwAuthenticatePresent == TRUE)
    {
    strcat(buffer, "WWW-Authenticate: ");
    CommaParam_List *params;
    if (msgHeader->wwwAuthenticate.challenge.present == TTCN_constant_Challenge_digestCln)
      {
      strcat(buffer, "Digest ");
      params = &msgHeader->wwwAuthenticate.challenge.__value.digestCln;
      }
    else
      {
      if (!strlen(msgHeader->wwwAuthenticate.challenge.__value.otherChallenge.authScheme))
        {
        strcat(buffer, msgHeader->wwwAuthenticate.challenge.__value.otherChallenge.authScheme);
        }
      params = &msgHeader->wwwAuthenticate.challenge.__value.otherChallenge.authParams;
      }
    int i;
    for (i=0; i<params->length; ++i)
      {
      if (((GenericParam *)params->elements)[i].idPresent == TRUE)
        {
        if ((i == 0 && !strlen(msgHeader->wwwAuthenticate.challenge.__value.otherChallenge.authScheme)) || i > 0)
          {
          strcat(buffer, ",");
          }
        strcat(buffer, ((GenericParam *)params->elements)[i].id);
        if (((GenericParam *)params->elements)[i].paramValuePresent == TRUE)
          {
          strcat(buffer, "=");
          strcat(buffer, ((GenericParam *)params->elements)[i].paramValue);
          }
        }
      }
    strcat(buffer, "\r\n");
    }
  }

static void d_sip_unknown(sip_t *sip, MessageHeader *msgHeader)
  {
  sip_unknown_t *header = sip->sip_unknown;
  msgHeader->undefinedHeader_ListPresent = (RTDS_BOOLEAN) (header != NULL);
  if (msgHeader->undefinedHeader_ListPresent == TRUE)
    {
    RTDS_UndefinedHeader_List_init(&(msgHeader->undefinedHeader_List), TRUE);
    int i = 0;
    while (header != NULL)
      {
      RTDS_UndefinedHeader_List_add_index(&(msgHeader->undefinedHeader_List), i, NULL);
      if (header->un_name != NULL)
        {
        strcpy(((UndefinedHeader *)msgHeader->undefinedHeader_List.elements)[i].headerName, header->un_name);
        }
      else
        {
        memset(((UndefinedHeader *)msgHeader->undefinedHeader_List.elements)[i].headerName, 0, sizeof(RTDS_String));
        }
      if (header->un_value != NULL)
        {
        strcpy(((UndefinedHeader *)msgHeader->undefinedHeader_List.elements)[i].headerValue, header->un_value);
        }
      else
        {
        memset(((UndefinedHeader *)msgHeader->undefinedHeader_List.elements)[i].headerValue, 0, sizeof(RTDS_String));
        }
      ++i;
      header = header->un_next;
      }
    }
  }

static void e_sip_unknown(char *buffer, MessageHeader *msgHeader)
  {
  if (msgHeader->undefinedHeader_ListPresent == TRUE)
    {
    int i;      
    for (i=0; i<msgHeader->undefinedHeader_List.length; ++i)
      {
      if (!strlen(((UndefinedHeader *)msgHeader->undefinedHeader_List.elements)[i].headerName))
        {
        strcat(buffer, ((UndefinedHeader *)msgHeader->undefinedHeader_List.elements)[i].headerName);
        strcat(buffer, ": ");
        if (!strlen(((UndefinedHeader *)msgHeader->undefinedHeader_List.elements)[i].headerValue))
          {
          strcat(buffer, ((UndefinedHeader *)msgHeader->undefinedHeader_List.elements)[i].headerValue);
          }
        strcat(buffer, "\r\n");
        }
      }
    }
  }

static void d_sip_header(sip_t *sip, MessageHeader *msgHeader)
  {  
  d_sip_accept(sip, msgHeader);
  d_sip_accept_encoding(sip, msgHeader);
  d_sip_accept_language(sip, msgHeader);
  d_sip_alert_info(sip, msgHeader);
  d_sip_allow(sip, msgHeader);
  d_sip_authentication_info(sip, msgHeader);
  d_sip_authorization(sip, msgHeader);
  d_sip_call_id(sip, msgHeader);
  d_sip_call_info(sip, msgHeader);
  d_sip_contact(sip, msgHeader);
  d_sip_content_disposition(sip, msgHeader);
  d_sip_content_encoding(sip, msgHeader);
  d_sip_content_language(sip, msgHeader);
  d_sip_content_length(sip, msgHeader);
  d_sip_content_type(sip, msgHeader);
  d_sip_cseq(sip, msgHeader);
  d_sip_date(sip, msgHeader);
  d_sip_error_info(sip, msgHeader);
  d_sip_expires(sip, msgHeader);
  d_sip_from(sip, msgHeader);
  d_sip_in_reply_to(sip, msgHeader);
  d_sip_max_forwards(sip, msgHeader);
  d_sip_mime_version(sip, msgHeader);
  d_sip_min_expires(sip, msgHeader);
  d_sip_organization(sip, msgHeader);
  d_sip_priority(sip, msgHeader);
  d_sip_proxy_authenticate(sip, msgHeader);
  d_sip_proxy_authorization(sip, msgHeader);
  d_sip_proxy_require(sip, msgHeader);
  d_sip_record_route(sip, msgHeader);
  d_sip_reply_to(sip, msgHeader);
  d_sip_require(sip, msgHeader);
  d_sip_retry_after(sip, msgHeader);
  d_sip_route(sip, msgHeader);
  d_sip_server(sip, msgHeader);
  d_sip_subject(sip, msgHeader);
  d_sip_supported(sip, msgHeader);
  d_sip_timestamp(sip, msgHeader);
  d_sip_to(sip, msgHeader);
  d_sip_unsupported(sip, msgHeader);
  d_sip_user_agent(sip, msgHeader);
  d_sip_via(sip, msgHeader); 
  d_sip_warning(sip, msgHeader);
  d_sip_www_authenticate(sip, msgHeader);
  d_sip_unknown(sip, msgHeader);
  }

static void e_sip_header(char *buffer, MessageHeader *msgHeader)
  {
  e_sip_accept(buffer, msgHeader);
  e_sip_accept_encoding(buffer, msgHeader);
  e_sip_accept_language(buffer, msgHeader);
  e_sip_alert_info(buffer, msgHeader);
  e_sip_allow(buffer, msgHeader);
  e_sip_authentication_info(buffer, msgHeader);
  e_sip_authorization(buffer, msgHeader);
  e_sip_call_id(buffer, msgHeader);
  e_sip_call_info(buffer, msgHeader);
  e_sip_contact(buffer, msgHeader);
  e_sip_content_disposition(buffer, msgHeader);
  e_sip_content_encoding(buffer, msgHeader);
  e_sip_content_language(buffer, msgHeader);
  e_sip_content_length(buffer, msgHeader);
  e_sip_content_type(buffer, msgHeader);
  e_sip_cseq(buffer, msgHeader);
  e_sip_date(buffer, msgHeader);
  e_sip_error_info(buffer, msgHeader);
  e_sip_expires(buffer, msgHeader);
  e_sip_from(buffer, msgHeader);
  e_sip_in_reply_to(buffer, msgHeader);
  e_sip_max_forwards(buffer, msgHeader);
  e_sip_mime_version(buffer, msgHeader);
  e_sip_min_expires(buffer, msgHeader);
  e_sip_organization(buffer, msgHeader);
  e_sip_priority(buffer, msgHeader);
  e_sip_proxy_authenticate(buffer, msgHeader);
  e_sip_proxy_authorization(buffer, msgHeader);
  e_sip_proxy_require(buffer, msgHeader);
  e_sip_record_route(buffer, msgHeader);
  e_sip_reply_to(buffer, msgHeader);
  e_sip_require(buffer, msgHeader);
  e_sip_retry_after(buffer, msgHeader);
  e_sip_route(buffer, msgHeader);
  e_sip_server(buffer, msgHeader);
  e_sip_subject(buffer, msgHeader);
  e_sip_supported(buffer, msgHeader);
  e_sip_timestamp(buffer, msgHeader);
  e_sip_to(buffer, msgHeader);
  e_sip_unsupported(buffer, msgHeader);
  e_sip_user_agent(buffer, msgHeader);
  e_sip_via(buffer, msgHeader);
  e_sip_warning(buffer, msgHeader);
  e_sip_www_authenticate(buffer, msgHeader);
  e_sip_unknown(buffer, msgHeader);
  }
  
static void d_sip_request_payload(sip_t *sip, Request *request)
  {
  sip_payload_t *payload = sip->sip_payload;
  request->messageBodyPresent = (RTDS_BOOLEAN) (payload != NULL);
  request->payloadPresent = request->messageBodyPresent;  
  if (request->payloadPresent == TRUE)
    {
    char *i = request->messageBody;
    request->payload.payloadlength = 0;
    char *j = request->payload.payloadvalue;
    while (payload != NULL)
      {
      request->payload.payloadlength += (int) payload->pl_len;
      if (payload->pl_data != NULL)
        {
        memcpy(i, payload->pl_data, payload->pl_len);
        i += payload->pl_len;
        memcpy(j, payload->pl_data, payload->pl_len);
        j += payload->pl_len;
        }
      payload = payload->pl_next;
      }
    request->messageBody[request->payload.payloadlength] = '\0';
    }
  }

static void e_sip_request_payload(char *buffer, Request *request)
  {
  strcat(buffer, "\r\n");
  if (request->messageBodyPresent == TRUE)
    {
    strcat(buffer, request->messageBody);
    }
  }

static void d_sip_response_payload(sip_t *sip, Response *response)
  {
  sip_payload_t *payload = sip->sip_payload;
  response->messageBodyPresent = (RTDS_BOOLEAN) (payload != NULL);
  response->payloadPresent = response->messageBodyPresent;  
  if (response->payloadPresent == TRUE)
    {
    char *i = response->messageBody;
    response->payload.payloadlength = 0;
    char *j = response->payload.payloadvalue;
    while (payload != NULL)
      {
      response->payload.payloadlength += (int) payload->pl_len;
      if (payload->pl_data != NULL)
        {
        memcpy(i, payload->pl_data, payload->pl_len);
        i += payload->pl_len;
        memcpy(j, payload->pl_data, payload->pl_len);
        j += payload->pl_len;
        }
      payload = payload->pl_next;
      }
    response->messageBody[response->payload.payloadlength] = '\0';
    }
  }

static void e_sip_response_payload(char *buffer, Response *response)
  {
  strcat(buffer, "\r\n");
  if (response->messageBodyPresent == TRUE)
    {
    strcat(buffer, response->messageBody);
    }
  }
  
Request *d_sip_request(char *buffer)
  {
  msg_t *msg = msg_make(sip_default_mclass(), 0, buffer, strlen(buffer)+1);
  sip_t *sip = sip_object(msg);
  Request *request = NULL;
  if (sip->sip_request != NULL)
    {
    request = (Request *) malloc(sizeof(Request));
    memset(request, 0, sizeof(Request));
    d_sip_request_line(sip, request);
    d_sip_header(sip, &request->msgHeader);
    d_sip_request_payload(sip, request);
    }
  msg_destroy(msg);
  return request;
  }

Response *d_sip_response(char *buffer)
  {
  msg_t *msg = msg_make(sip_default_mclass(), 0, buffer, strlen(buffer)+1);
  sip_t *sip = sip_object(msg);
  Response *response = NULL;
  if (sip->sip_status != NULL)
    {
    response = (Response *) malloc(sizeof(Response));
    memset(response, 0, sizeof(Response));
    d_sip_status_line(sip, response);
    d_sip_header(sip, &response->msgHeader);
    d_sip_response_payload(sip, response);
    }
  msg_destroy(msg);
  return response;
  }

void e_sip_request(char *buffer, Request *request)
  {
  e_sip_request_line(buffer, request);
  e_sip_header(buffer, &request->msgHeader);
  e_sip_request_payload(buffer, request);
  }

void e_sip_response(char *buffer, Response *response)
  {
  e_sip_status_line(buffer, response);
  e_sip_header(buffer, &response->msgHeader);
  e_sip_response_payload(buffer, response);
  }

void SUT_send(RTDS_MessageHeader *msg, address *addr, int portIndex)
  {
  /* Set address to send */
  struct sockaddr_in addr_in;
  memset(&addr_in, 0, sizeof(addr_in));
  addr_in.sin_family = AF_INET;
  /* If address is given, use it */
  if (addr != NULL)
    {
    addr_in.sin_addr.s_addr = inet_addr(addr->host);
    addr_in.sin_port = htons(addr->portField);
    }
  /* If no address given, use the default one */
  else
    {
    addr_in.sin_addr.s_addr = inet_addr(DEFAULT_SUT_ADDRESS);
    addr_in.sin_port = htons(DEFAULT_SUT_PORT);
    }
  /* Encode message */
  char buffer[BUFFER_SIZE];
  switch (msg->messageNumber)
    {
    case RTDS_message_Request:
    case RTDS_message_REGISTER_Request:
    case RTDS_message_INVITE_Request:
    case RTDS_message_OPTIONS_Request:
    case RTDS_message_BYE_Request:
    case RTDS_message_CANCEL_Request:
    case RTDS_message_ACK_Request:
      e_sip_request(buffer, (Request *) msg->pData);
      break;
    case RTDS_message_Response:
      e_sip_response(buffer, (Response *) msg->pData);
      break;
    case RTDS_message_Raw:
    case RTDS_message_Raw_REGISTER_Request:
    case RTDS_message_Raw_INVITE_Request:
    case RTDS_message_Raw_BYE_Request:
    case RTDS_message_Raw_Response:
    case RTDS_message_Raw_Unknown_Request:
      memcpy(buffer, msg->pData, msg->dataLength);
      buffer[msg->dataLength] = '\0';
      break;
    }
  /* Send message via given port */
  sendto(SUT_sockets[portIndex], buffer, strlen(buffer) + 1, 0, (struct sockaddr *) &addr_in, sizeof(addr_in));
  }

#ifdef WIN32
DWORD WINAPI SUT_recv(void *arg)
#else
void *SUT_recv(void *arg)
#endif
  {
  int portIndex = *((int *) arg);
  char buffer[BUFFER_SIZE];
  /* Wait indefinitely for messages */
  for (;;)
    {
    /* Receive data */
    memset(buffer, 0, BUFFER_SIZE);
    struct sockaddr_in addr_in;
#ifdef WIN32
    int addr_len = (int) sizeof(addr_in);
#else
    socklen_t addr_len = (int) sizeof(addr_in);
#endif
    memset(&addr_in, 0, addr_len);
    if (!(recvfrom(SUT_sockets[portIndex], buffer, BUFFER_SIZE, 0, (struct sockaddr *) &addr_in, &addr_len) > 0))
      {
      break;
      }
    /* Get sender address */
    address addr;
    strcpy(addr.host, inet_ntoa(addr_in.sin_addr));
    addr.portField = ntohs(addr_in.sin_port);
    /* Message is a request */
    Request *request = d_sip_request(buffer);
    if (request != NULL)
      {
      switch (request->requestLine.method)
        {
        case TTCN_constant_ACK_E:
          RTDS_MSG_QUEUE_SEND_TO_QUEUE_ID_FROM_SUT_WITH_ADDRESS(
            RTDS_message_ACK_Request, 
            sizeof(Request), 
            (unsigned char *) request, 
            NULL, 
            SUT_systemPorts[portIndex].portMessageQueue, 
            (&addr));
          break;
        case TTCN_constant_BYE_E:
          RTDS_MSG_QUEUE_SEND_TO_QUEUE_ID_FROM_SUT_WITH_ADDRESS(
            RTDS_message_BYE_Request, 
            sizeof(Request), 
            (unsigned char *) request, 
            NULL, 
            SUT_systemPorts[portIndex].portMessageQueue, 
            (&addr));
          break;
        case TTCN_constant_CANCEL_E:
          RTDS_MSG_QUEUE_SEND_TO_QUEUE_ID_FROM_SUT_WITH_ADDRESS(
            RTDS_message_CANCEL_Request, 
            sizeof(Request), 
            (unsigned char *) request, 
            NULL, 
            SUT_systemPorts[portIndex].portMessageQueue, 
            (&addr));
          break;
        case TTCN_constant_INVITE_E:
          RTDS_MSG_QUEUE_SEND_TO_QUEUE_ID_FROM_SUT_WITH_ADDRESS(
            RTDS_message_INVITE_Request, 
            sizeof(Request), 
            (unsigned char *) request, 
            NULL, 
            SUT_systemPorts[portIndex].portMessageQueue, 
            (&addr));
          break;
        case TTCN_constant_OPTIONS_E:
          RTDS_MSG_QUEUE_SEND_TO_QUEUE_ID_FROM_SUT_WITH_ADDRESS(
            RTDS_message_OPTIONS_Request, 
            sizeof(Request), 
            (unsigned char *) request, 
            NULL, 
            SUT_systemPorts[portIndex].portMessageQueue, 
            (&addr));
          break;
        case TTCN_constant_REGISTER_E:
          RTDS_MSG_QUEUE_SEND_TO_QUEUE_ID_FROM_SUT_WITH_ADDRESS(
            RTDS_message_REGISTER_Request, 
            sizeof(Request), 
            (unsigned char *) request, 
            NULL, 
            SUT_systemPorts[portIndex].portMessageQueue, 
            (&addr));
          break;
        default:
          RTDS_MSG_QUEUE_SEND_TO_QUEUE_ID_FROM_SUT_WITH_ADDRESS(
            RTDS_message_Request, 
            sizeof(Request), 
            (unsigned char *) request, 
            NULL, 
            SUT_systemPorts[portIndex].portMessageQueue, 
            (&addr));
          break;
        }
      continue;
      }
    /* Message is a response */
    Response *response = d_sip_response(buffer);
    if (response != NULL)
      {
      RTDS_MSG_QUEUE_SEND_TO_QUEUE_ID_FROM_SUT_WITH_ADDRESS(
        RTDS_message_Response, 
        sizeof(Response), 
        (unsigned char *) response, 
        NULL, 
        SUT_systemPorts[portIndex].portMessageQueue, 
        (&addr));
      }
    }
    
  free(arg);
#ifdef WIN32
  return 0;
#else
  return NULL;
#endif
  }

int SUT_init(void)
  {
#ifdef WIN32
  WSADATA wsaData;
  WSAStartup(MAKEWORD(2, 2), &wsaData);
#endif
  /* Setup system interface ports */
  int i;
  for (i=UDP1; i<=UDP3; ++i)
    {
    SUT_systemPorts[i].portName = SUT_portNames[i];
    SUT_systemPorts[i].portMessageQueue = RTDS_NEW_MESSAGE_QUEUE;
    if (i != UDP3)
      {
      SUT_systemPorts[i].next = &SUT_systemPorts[i+1];
      }
    }
  SUT_systemPorts[UDP3].next = NULL;
  RTDS_TTCN_systemInterfacePorts = SUT_systemPorts;
  /* Setup sockets */
  for (i=UDP1; i<=UDP3; ++i)
    {
    /* Create socket */
    SUT_sockets[i] = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (IS_INVALID_SOCKET(SUT_sockets[i]))
      {
      fprintf(stderr, "%d: Error creating socket!\n", SOCKET_ERROR_NUMBER);
      return -1;
      }
    /* Make socket reusable */
    int yes = 1;
    if (IS_SOCKET_ERROR(setsockopt(SUT_sockets[i], SOL_SOCKET, SO_REUSEADDR, (const char*) &yes, sizeof(yes))))
      {
      fprintf(stderr, "%d: Error making socket reusable!\n", SOCKET_ERROR_NUMBER);
      closesocket(SUT_sockets[i]);
      SUT_sockets[i] = INVALID_SOCKET;
      return -1;
      }
    /* Setup address */
    struct sockaddr_in addr_in;
    memset(&addr_in, 0, sizeof(addr_in));
    addr_in.sin_family = AF_INET;
    addr_in.sin_addr.s_addr = INADDR_ANY;//inet_addr(DEFAULT_SUT_ADDRESS);
    addr_in.sin_port = htons(SUT_portNumbers[i]);
    /* Bind to address */
    if (IS_SOCKET_ERROR(bind(SUT_sockets[i], (struct sockaddr *) &addr_in, sizeof(addr_in))))
      {
      fprintf(stderr, "%d: Error binding socket!\n", SOCKET_ERROR_NUMBER);
      closesocket(SUT_sockets[i]);
      SUT_sockets[i] = INVALID_SOCKET;
      return -1;
      }
    /* Create receiving thread */
    int *arg = (int *) malloc(sizeof(int));
    *arg = i;
#ifdef WIN32
    DWORD threadId;
    SUT_thread[i] = CreateThread(NULL, 0, &SUT_recv, (void *) arg, 0, &threadId);
    if (SUT_thread == NULL)
#else
    if (pthread_create(&SUT_thread[i], NULL, SUT_recv, (void *) arg))
#endif
      {
      fprintf(stderr, "Error creating receiving thread!\n");
      closesocket(SUT_sockets[i]);
      SUT_sockets[i] = INVALID_SOCKET;
      return -1;
      }
    } 
  return 0;
  }

int SUT_clean(void)
  {
  int i;
  for (i=UDP1; i<=UDP3; ++i)
    {
    shutdown(SUT_sockets[i], 2);
    closesocket(SUT_sockets[i]);
#ifdef WIN32
    WaitForSingleObject(SUT_thread[i], INFINITE);
    CloseHandle(SUT_thread[i]);
#else
    pthread_join(SUT_thread[i], NULL);
#endif
    }
#ifdef WIN32
  WSACleanup();
#endif
  return 0;
  }
  
